/**
  *  AUTOLOG.g4  
  *  AutoLog programming language grammar, 
  *  coherent logic forms, Type terms/term types, 
  *  impredicativity, functorials, modulators.
  *  ANTLR4 tools.
  *  @version 4/10/2021 (default DUOP)
  *  @author &copy; john.r.fisher@protonmail.com
     */
grammar AUTOLOG ;

@header{
package sam.antlr.autolog ; 
}

theory : inference* ; // inference = rule or modulator

inference : 
   conjunction INFER disjunction '.'   
                             # inference_rule
   |  term EQ term '.'       # inference_rewrite
   ;
disjunction : 
   conjunction (('|' | ';') conjunction)* 
   ; 
conjunction : 
   literal (',' literal)*
   ;
literal : // always translate as Functor
   // impredicative prop
   VARIABLE                  # literal_impprop    
   // terms other than impprop
   | term                    # literal_term       
   ; 
term :    // earlier case forces tighter binding       
   NUMBER                   # term_number
   | VARIABLE               # term_variable 
   | name                   # term_name
   | term '(' terms ')'     # term_functor // ALLOWS PARAMETRIC 
   | POSSIBLE term          # possible_term 
   | NECESSARY term         # necessary_term
   | GEN term               # gen_term
   | NEG  term              # term_neg
   | LNEG term              # term_lneg
   | MINUS term             # term_minus
   | SHARP term             # term_sharp
   | term STAR term         # term_star
   | term CIRCLE term       # term_circle
   | term CCIRCLE term      # term_ccircle
   | term AND term          # term_and
   | term CCFLEX term       # term_ccflex
   | term MEET term         # term_meet
   | term INTERSECT term    # term_intersect
   | term JOIN term         # term_join
   | term UNION term        # term_union
   | term PLUS term         # term_plus
   |<assoc=right>          // right associative
     term IF term           # term_if
   |<assoc=right>          // right associative          
     term ARROW term        # term_arrow  
   |<assoc=right>          // right associative          
     term BSLASH term       # term_bslash  
   | term FI term           # term_fi
   | term LARROW term       # term_larrow  
   | term FSLASH term       # term_fslash
   | term MEMBER term       # term_member
   | term EQ term           # term_eq  
   | term IFF term          # term_iff
   | term EQUIV term        # term_equiv
   | term NEQ term          # term_neq  
   | term LESS term         # term_less
   | term LESSEQ term       # term_lesseq
   | term GREATER term      # term_greater
   | term GREATEREQ term    # term_greatereq
   | '(' t=term ')'         # term_paren
   | term ':' term          # term_judgement  
   /////  DUOPs dynamic user operators (copy/pasted)
   | DUOP '(' terms ')'     # term_functional_DUOP
   | DUOP term              # term_prefix_DUOP   // pre << post << in
   | term DUOP term         # term_infix_DUOP 
   /////  end DUOPs  
   | '{' term '|' term '}'  # term_collection  // Functor type 4
   | '(' term ',' term ')'  # term_pair        // Functor type 5
   ;
terms :  
   t=term (',' t=term )* 
   ; 
name :  
   ALL|SOME|LAMBDA|PI|SUM|TOP|BOTTOM|POSSIBLE|NECESSARY 
   |GEN|NEG|LNEG|EQ|IFF|EQUIV|NEQ|STAR|CIRCLE|CCIRCLE|AND 
   |CCFLEX|MEET|INTERSECT|JOIN|UNION|MEMBER|EMPTY|PLUS|MINUS 
   |IF|BSLASH|FI|FSLASH|ARROW|LARROW|SHARP|LESS|LESSEQ 
   |GREATER|GREATEREQ|LOWER|UPPER|NUMERIC|QUOTED|DUOP|LNAME 
   ; 
LNAME :  // lexed general name
   LOWER (LOWER | UPPER | '_' | '-' | NUMERIC )* 
   ;
NUMBER :   
   NUMERIC+
   | NUMERIC+ '.' NUMERIC+ 
   ;
VARIABLE : // similar to Prolog
   (UPPER | '_' | '?')  (LOWER | UPPER | '_' | NUMERIC)* /
   ; 
ALL :     '∀' ; // logic forall  
SOME :    '∃' ; // logic forsome 
LAMBDA :  'λ' ; // Greek lambda  
PI :      'Π' ; // Greek Pi      
SUM :     'Σ' ; // Greek Sigma   
TOP :     '⊤' ; // top, truth      
BOTTOM :  '⊥' ; // bottom, falsum  
POSSIBLE : '◇'  ; // modal,  cta (click edit button to add)
NECESSARY : '□' ; // modal cta
GEN :     '@'  ;
NEG :     '~'  ;  
LNEG :    '¬'  ; // logic negation, option l 
EQ :      '='  ; 
IFF :     '↔' ; // cta 
EQUIV :   '≈' ; // option x 
NEQ :     '≠'  ; // not equal, option =  
STAR :    '*'  ;
CIRCLE :  '∘'  ;  // cta  
CCIRCLE : '∙'  ;  // cta  
AND :     '&'  ;
CCFLEX :  '^'  ;  // shft+6, e.g. 2^3 (=8)
MEET :    '∧'  ;  // cta   
INTERSECT : '∩'  ;  // cta 
JOIN :    '∨'  ;  // cta   
UNION :   '∪'  ;  // cta   
MEMBER :  '∈'  ;  // cta   
EMPTY :   '∅'  ;  // cta  
PLUS :    '+'  ;
MINUS :   '-'  ;
IF :      '→'  ;  // cta 
BSLASH:   '\\' ;  // backslash
FI :      '←'  ;  // cta 
FSLASH:   '/' ;
ARROW :   '->' ;
LARROW :  '<-' ;
SHARP :   '#'  ;
LESS  :   '<'  ;
LESSEQ :  '≤'  ;  // option <
GREATER : '>'  ;
GREATEREQ:'≥'  ;  // option >
LOWER :   'a'..'z'    ; 
UPPER :   'A'..'Z'    ; 
NUMERIC : '0'..'9'    ; 
QUOTED :    // embed double quoted "data"
   // e.g., "This is a comment."
   '\u{0022}' .*? '\u{0022}'  // "any" charseq
   ;
DUOP :  // Dynamic User Operators  
        // A- copy these from table and paste in editor
    // arrows
          '↠'|'↰'|'⇀'|'⇐'|'⇠'|'⇰'
     |'↑'|'↡'|'↱'|'⇁'|'⇑'|'⇡'|'⇱'
     |    '↢'|'↲'|'⇂'|'⇒'|'⇢'|'⇲'
     |'↓'|'↣'|'↳'|'⇃'|'⇓'|'⇣'|'⇳'
     |    '↤'|'↴'|'⇄'|'⇔'|'⇤'|'⇴'
     |'↕'|'↥'|'↵'|'⇅'|'⇕'|'⇥'|'⇵'
     |'↖'|'↦'|'↶'|'⇆'|'⇖'|'⇦'|'⇶'
     |'↗'|'↧'|'↷'|'⇇'|'⇗'|'⇧'|'⇷'
     |'↘'|'↨'|'↸'|'⇈'|'⇘'|'⇨'|'⇸'
     |'↙'|'↩'|'↹'|'⇉'|'⇙'|'⇩'|'⇹'
     |'↚'|'↪'|'↺'|'⇊'|'⇚'|'⇪'|'⇺'
     |'↛'|'↫'|'↻'|'⇋'|'⇛'|'⇫'|'⇻'
     |'↜'|'↬'|'↼'|'⇌'|'⇜'|'⇬'|'⇼'
     |'↝'|'↭'|'↽'|'⇍'|'⇝'|'⇭'|'⇽'
     |'↞'|'↮'|'↾'|'⇎'|'⇞'|'⇮'|'⇾'
     |'↟'|'↯'|'↿'|'⇏'|'⇟'|'⇯'|'⇿'
    // greek
     |'α'|'β'|'γ'|'δ'|'ε'|'ζ'|'η'|'θ'|'ι'|'κ'|    'μ'|'ν'
     |'Α'|'Β'|'Γ'|'Δ'|'Ε'|'Ζ'|'Η'|'Θ'|'Ι'|'Κ'|'Λ'|'Μ'|'Ν'
     |'ξ'|'ο'|'π'|'ρ'|'ς'|'σ'|'τ'|'υ'|'φ'|'χ'|'ψ'|'ω'
     |'Ξ'|'Ο'|    'Ρ'|    'Τ'|'Υ'|'Φ'|'Χ'|'Ψ'|'Ω'
    // math ops  
  |'∐'|'∠'|'∮∮∮'|'≀'|'≐'|     '≰'|'⊀'|'⊐'|'⊠'|'⊰'|'⋀'|'⋐'|'⋠'|'⋰'
  |'∁'|'∑'|'∡'|'∱'|'≁'|'≑'|'≡'|'≱'|'⊁'|'⊑'|'⊡'|'⊱'|'⋁'|'⋑'|'⋡'|'⋱'
  |'∂'|'−'|'∢'|'∲'|'≂'|'≒'|'≢'|'≲'|'⊂'|'⊒'|'⊢'|'⊲'|'⋂'|'⋒'|'⋢'|'⋲'
      |'∓'|'∣'|'∳'|'≃'|'≓'|'≣'|'≳'|'⊃'|'⊓'|'⊣'|'⊳'|'⋃'|'⋓'|'⋣'|'⋳'
  |'∄'|'∔'|'∤'|'∴'|'≄'|'≔'|    '≴'|'⊄'|'⊔'|    '⊴'|'⋄'|'⋔'|'⋤'|'⋴'
  |    '∕'|'∥'|'∵'|'≅'|'≕'|    '≵'|'⊅'|'⊕'|    '⊵'|'⋅'|'⋕'|'⋥'|'⋵'
  |'∆'|'∖'|'∦'|'∶'|'≆'|'≖'|'≦'|'≶'|'⊆'|'⊖'|'⊦'|'⊶'|'⋆'|'⋖'|'⋦'|'⋶'
  |'∇'|'∗'|    '∷'|'≇'|'≗'|'≧'|'≷'|'⊇'|'⊗'|'⊧'|'⊷'|'⋇'|'⋗'|'⋧'|'⋷'
  |            '∸'|    '≘'|'≨'|'≸'|'⊈'|'⊘'|'⊨'|'⊸'|'⋈'|'⋘'|'⋨'|'⋸'
  |'∉'|        '∹'|'≉'|'≙'|'≩'|'≹'|'⊉'|'⊙'|'⊩'|'⊹'|'⋉'|'⋙'|'⋩'|'⋹'
  |'∊'|'√'|    '∺'|'≊'|'≚'|'≪'|'≺'|'⊊'|'⊚'|'⊪'|'⊺'|'⋊'|'⋚'|'⋪'|'⋺'
  |'∋'|'∛'|'∫'|'∻'|'≋'|'≛'|'≫'|'≻'|'⊋'|'⊛'|'⊫'|'⊻'|'⋋'|'⋛'|'⋫'|'⋻'
  |'∌'|'∜'|'∫∫'|'∼'|'≌'|'≜'|'≬'|'≼'|'⊌'|'⊜'|'⊬'|'⊼'|'⋌'|'⋜'|'⋬'|'⋼'
  |'∍'|'∝'|'∫∫∫'|'∽'|'≍'|'≝'|'≭'|'≽'|'⊍'|'⊝'|'⊭'|'⊽'|'⋍'|'⋝'|'⋭'|'⋽'
  |'∎'|'∞'|'∮'|'∾'|'≎'|'≞'|'≮'|'≾'|'⊎'|'⊞'|'⊮'|'⊾'|'⋎'|'⋞'|'⋮'|'⋾'
  |'∏'|'∟'|'∮∮'|'∿'|'≏'|'≟'|'≯'|'≿'|'⊏'|'⊟'|'⊯'|'⊿'|'⋏'|'⋟'|'⋯'|'⋿'
     |'–'      //  keyboard option -   (subtraction)
     |'⨯'      //  cross product 
     | '\'' .*? '\''            // 'any' single-quote charsequence
     ;
WS : [ \t\n\r]+ -> skip ; // ignore all whitespace 
SL_COMMENT : ('%'|'//') .*? ('\n'|'\r') -> skip ;
ML_COMMENT : ('/*' | '(*') .*? ('*/' | '*)' )  -> skip ; 
INFER   : '=>' ; 
// REWRITE : '≈>' ;  // directional =   ??? not used


/*
(name cannot be lex)
8/24 VERSION NOTES 
- DUOPs are "meta" ops having less tight parse binding because
  they try to parse later in the grammar rule order.  The higher rules
  parse more tightly because they get to bind before lower rules can try,
  unless parens hold terms together.
- Note that parametric functors MUST have TYPE FUNCTOR.FUNCTIONAL == (byte)0.
- We are abandoning parse for postfix ops, for now.
- #term_collection is ako set builder
*/
