package sam.antlr.autolog ;

import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.tree.ErrorNode;
import org.antlr.v4.runtime.tree.TerminalNode;

import java.util.* ;
import sam.machine.lang.*  ;   // machine lang translation target

 /**
  *  AUTOLOG Reader
  *  implements the AUTOLOGBaseListener walking methods,
  *  and builds sam.machine.lang components during a walk.
  *  @version 10/11/2020  (collection & pair terms)
  *  { Try to specify #methods in AUTOLOG.g4 grammar-rule #order. }
   */
public class Reader extends AUTOLOGBaseListener {  
   // stack helps build literals, terms, and conjunctions
   Stack<Term> stack  ;  
   //  Making Rules 
   public Theory theory ; 
   Conjunction antecedent ;  // for inference rule 
   Disjunction consequent ;  // for inference rule

/////-----
   public void enterTheory(AUTOLOGParser.TheoryContext ctx) { 
      stack =  new Stack<Term>() ; 
      theory = new Theory() ;
   }
/////----- # inference_rule 
   public void enterInference_rule(AUTOLOGParser.Inference_ruleContext ctx) {
      // initialize 
      antecedent = null ;  // signal that antecedent not yet built
   }
   public void exitInference_rule(AUTOLOGParser.Inference_ruleContext ctx) { 
      theory.add(new Rule(antecedent,consequent)) ; 
   }
/////----- # inference_rewrite
   public void exitInference_rewrite(AUTOLOGParser.Inference_rewriteContext ctx) {
      Term Bterm = stack.pop() ; 
      Term Aterm = stack.pop() ;
      theory.add(new Rewrite(Aterm,Bterm)) ;  // Aterm -> Bterm .
   } 
/////-----
   public void enterDisjunction(AUTOLOGParser.DisjunctionContext ctx) {
      consequent = new Disjunction() ; 
   }
/////-----
   public void exitConjunction(AUTOLOGParser.ConjunctionContext ctx) { 
      if (antecedent == null) {  
         antecedent = new Conjunction() ; 
         while(!stack.isEmpty())  antecedent.add(0,(Functor)stack.pop()) ;
      }
      else {  
         Conjunction c = new Conjunction() ;
         while(!stack.isEmpty()) c.add(0,(Functor)stack.pop()) ; 
         // stack now empty
         consequent.add(c) ;
      }
   }
/////----- # literal_impprop
   public void exitLiteral_impprop(AUTOLOGParser.Literal_imppropContext ctx) { 
      Variable v = new Variable(ctx.VARIABLE().getText()) ;
      Functor f =  new Functor("",v,null,(byte)0) ;  // make fnctr out of v
      //??//                   ^
      stack.push(f) ;  // variable --> literal
   }
/////----- # literal_term
   public void exitLiteral_term(AUTOLOGParser.Literal_termContext ctx) { 
      // perhaps check stack.peek() instanceof Functor 
      // Term t = stack.pop() ;  // term created
      // stack.push(t) ;   
      // i.e., leave term on the stack ;-)
   }
/////----- # term_number
   public void exitTerm_number(AUTOLOGParser.Term_numberContext ctx) {  
      Functor num = new Functor(ctx.getText(),null,null,(byte)0) ;
      num.numeric = true ;
      stack.push(num) ;
   }
/////----- # term_variable
   public void exitTerm_variable(AUTOLOGParser.Term_variableContext ctx) {  
      Variable v = new Variable(ctx.VARIABLE().getText()) ;
      stack.push(v) ;   
   }
/////----- # term_name
   public void exitTerm_name(AUTOLOGParser.Term_nameContext ctx) { 
      stack.push(new Functor(ctx.getText(),null,null, (byte)0)) ;
   }
/////----- # term_functor   captures parametric functors
   //  term: ...  |  term   '(' terms ')'  # term_functor
   //  "t(t1,...,tn)" will parse to stack=[tn,...,t2,t], with tn on top.
   public void exitTerm_functor(AUTOLOGParser.Term_functorContext ctx)  { 
      int n = ctx.terms().term().size() ;  // how many terms
      Term[] arguments = new Term[n] ; 
      // pop args from stack
      for (int i = n-1 ; 0 <= i ; i--) arguments[i] = stack.pop() ;  // reverse
      Term term = stack.pop() ; 
      Functor f ; 
      // term is Variable, e.g., P(...)
      if (term instanceof Variable)  
         f = new Functor("",term,arguments,(byte)0) ;
      // term is name or built-in
      else if (term instanceof Functor && ((Functor)term).args==null) 
         f = new Functor(((Functor)term).name,null,arguments,(byte)0) ; 
      // term is functorial
      else  
         f = new Functor("",term,arguments,(byte)0) ; 
      stack.push(f) ;    
   }
/////----- # possible_term
   public void exitPossible_term(AUTOLOGParser.Possible_termContext ctx) { 
      Term[] args = new Term[1] ; 
      args[0] = stack.pop() ;
      stack.push(new Functor("◇",null,args,(byte)1)) ;
   }
/////----- # necessary_term
   public void exitNecessary_term(AUTOLOGParser.Necessary_termContext ctx) { 
      Term[] args = new Term[1] ; 
      args[0] = stack.pop() ;
      stack.push(new Functor("□",null,args,(byte)1)) ;
   }
/////----- # gen_term
   public void exitGen_term(AUTOLOGParser.Gen_termContext ctx) { 
      Term[] args = new Term[1] ; 
      args[0] = stack.pop() ;
      stack.push(new Functor("@",null,args,(byte)1)) ;
   }
/////----- # term_neg
   public void exitTerm_neg(AUTOLOGParser.Term_negContext ctx) { 
      Term[] args = new Term[1] ; 
      args[0] = stack.pop() ;
      stack.push( new Functor("~",null,args,(byte)1)) ;
   }
/////----- # term_lneg
   public void exitTerm_lneg(AUTOLOGParser.Term_lnegContext ctx) { 
      Term[] args = new Term[1] ; 
      args[0] = stack.pop() ;
      stack.push( new Functor("¬",null,args,(byte)1)) ;
   }
/////----- # term_minus
   public void exitTerm_minus(AUTOLOGParser.Term_minusContext ctx) { 
      Term[] args = new Term[1] ; 
      args[0] = stack.pop() ;
      stack.push(new Functor("-",null,args,(byte)1)) ;
   }
/////----- # term_sharp
   public void exitTerm_sharp(AUTOLOGParser.Term_sharpContext ctx) { 
      Term[] args = new Term[1] ; 
      args[0] = stack.pop() ;
      stack.push(new Functor("#",null,args,(byte)1)) ;
   }
/////----- # term_star
   public void exitTerm_star(AUTOLOGParser.Term_starContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("*",null,args,(byte)2)) ;
   } 
/////----- # term_circle
   public void exitTerm_circle(AUTOLOGParser.Term_circleContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("∘",null,args,(byte)2)) ;
   } 
/////----- # term_ccircle
   public void exitTerm_ccircle(AUTOLOGParser.Term_ccircleContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("∙",null,args,(byte)2)) ;
   } 
/////----- # term_and
   public void exitTerm_and(AUTOLOGParser.Term_andContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("&",null,args,(byte)2)) ;
   } 
/////----- # term_ccflex
   public void exitTerm_ccflex(AUTOLOGParser.Term_ccflexContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("^",null,args,(byte)2)) ;
   } 
/////----- # term_meet
   public void exitTerm_meet(AUTOLOGParser.Term_meetContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("∧",null,args,(byte)2)) ;
   } 
/////----- # term_intersect
   public void exitTerm_intersect(AUTOLOGParser.Term_intersectContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("∩",null,args,(byte)2)) ;
   } 
/////----- # term_join
   public void exitTerm_join(AUTOLOGParser.Term_joinContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("∨",null,args,(byte)2)) ;
   } 
/////----- # term_union
   public void exitTerm_union(AUTOLOGParser.Term_unionContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("∪",null,args,(byte)2)) ;
   } 
/////----- # term_plus
   public void exitTerm_plus(AUTOLOGParser.Term_plusContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("+",null,args,(byte)2)) ;
   }
/////----- # term_if
   public void exitTerm_if(AUTOLOGParser.Term_ifContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("→",null,args,(byte)2)) ;
   }
/////----- # term_arrow
   public void exitTerm_arrow(AUTOLOGParser.Term_arrowContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("->",null,args,(byte)2)) ;
   }
/////----- # term_bslash
   public void exitTerm_bslash(AUTOLOGParser.Term_bslashContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("\\",null,args,(byte)2)) ;
   }
/////----- # term_fi
   public void exitTerm_fi(AUTOLOGParser.Term_fiContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("←",null,args,(byte)2)) ;
   }
/////----- # term_larrow
   public void exitTerm_larrow(AUTOLOGParser.Term_larrowContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("<-",null,args,(byte)2)) ;
   }
/////----- # term_fslash
   public void exitTerm_fslash(AUTOLOGParser.Term_fslashContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("/",null,args,(byte)2)) ;
   }
/////----- # term_member
   public void exitTerm_member(AUTOLOGParser.Term_memberContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("∈",null,args,(byte)2)) ;
   }
/////----- # term_eq
   public void exitTerm_eq(AUTOLOGParser.Term_eqContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("=",null,args,(byte)2)) ;
   }
/////----- # term_iff
   public void exitTerm_iff(AUTOLOGParser.Term_iffContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("↔",null,args,(byte)2)) ;
   }
/////----- # term_equiv
   public void exitTerm_equiv(AUTOLOGParser.Term_equivContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("≈",null,args,(byte)2)) ;
   }
/////----- # term_neq
   public void exitTerm_neq(AUTOLOGParser.Term_neqContext ctx) {
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("≠",null,args,(byte)2)) ;
   }
/////----- # term_less
   public void exitTerm_less(AUTOLOGParser.Term_lessContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("<",null,args,(byte)2)) ;
   }
/////----- # term_lesseq
   public void exitTerm_lesseq(AUTOLOGParser.Term_lesseqContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("≤",null,args,(byte)2)) ;
   }
/////----- # term_greater
   public void exitTerm_greater(AUTOLOGParser.Term_greaterContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor(">",null,args,(byte)2)) ;
   }
/////----- # term_greatereq
   public void exitTerm_greatereq(AUTOLOGParser.Term_greatereqContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("≥",null,args,(byte)2)) ;
   } 
//////////---------- # term_paren
   // NO method required: grammar is grouping subterms with parens
/////----- # term_judgement
   public void exitTerm_judgement(AUTOLOGParser.Term_judgementContext ctx) {
      Term right = stack.pop() ,   // last pushed
           left =  stack.pop() ; 
      Term[] args = new Term[2] ; 
      args[0] = left ;
      args[1] = right ;
      Functor f = new Functor(":",null,args,(byte)2) ; 
      stack.push(f) ;
   }
   public void exitTerm_functional_DUOP(AUTOLOGParser.Term_functional_DUOPContext ctx) {  
      int n = ctx.terms().term().size() ;  // how many terms
      Term[] args = new Term[n] ; 
      for (int i = n-1 ; 0 <= i ; i--) args[i] = stack.pop() ;  // reverse
      Functor f = new Functor(ctx.DUOP().getText(),null,args,(byte)0) ; // FUNCTIONAL
      stack.push(f) ;
   }
/////----- # term_prefix_DUOP
   public void exitTerm_prefix_DUOP(AUTOLOGParser.Term_prefix_DUOPContext ctx) { 
      Term[] args = new Term[1] ;  
      args[0] = stack.pop() ;
      stack.push(new Functor(ctx.DUOP().getText(),null,args,(byte)1)) ; //PREFIX
   }
/////----- # term_infix_DUOP
   public void exitTerm_infix_DUOP(AUTOLOGParser.Term_infix_DUOPContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor(ctx.DUOP().getText(),null,args,(byte)2)) ; // INFIX
   } 
/////----- # term_collection
   public void exitTerm_collection(AUTOLOGParser.Term_collectionContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("{-|-}",null,args,(byte)4)) ; // type 4 
   } 
/////----- # term_pair
   public void exitTerm_pair(AUTOLOGParser.Term_pairContext ctx) { 
      Term[] args = new Term[2] ; 
      args[1] = stack.pop() ;  
      args[0] = stack.pop() ;
      stack.push(new Functor("(-,-)",null,args,(byte)5)) ; // type 5
   } 

} // end Reader


