package sam.machine.active ;

import sam.machine.lang.* ;
import java.util.* ; 

/**   
  * ActiveTheory determined by a sam.machine.lang.Theory
  * @version 10/26/2023   
  *    ActiveTree, ActiveTheory, ActiveRule, ActiveRewrite 
  *    all SHARE instance variable theory, and thus can
  *    access AA data via reference "this.theory". 
  *    ... continue at line #77 with a report() 
  *    ... assertDynamic method should avoid repeats
  *    ... each inference stage should saturate inferences  
  *        for definite and dynamic inferences and then attempt a
  *        limited existential inference --- TBD  
   */
public class ActiveTheory  { 
   public ActiveTree tree ;   // the inference tree manager
   public Theory theory ;     // the source program theory
   public int initialSize ;   // the static theory size

   /// compiled active program ////////////////////////////////////////
   // definite rule or predicative rewrite
   protected Vector<ActiveInference> definite ;     
   // existential ActiveRule(s), use theory order
   protected Vector<ActiveInference> existential ;  
   // dynamic ActiveRewrite(s), asserted for new ActiveRule consequent fact
   protected Vector<ActiveRewrite> dynamic ; 
   ////////////////////////////////////////////////////////////////////

   /**
     *  Construct initial ActiveTheory from source program. 
     *  InferenceManager apply active inferences in stages:
     *    - Saturate active definite rules and active rewrites
     *    - then apply an existential rule in a limited manner.
      */
   public ActiveTheory(ActiveTree tree) { 
      this.tree = tree ;   
      this.theory = tree.theory ;
      this.initialSize = this.theory.size() ;   
      int indx = 1 ;   
      Iterator it = theory.iterator() ; 
      while (it.hasNext()) { 
         Object next = it.next() ;
         if (next instanceof Rule) {   
            Rule rule = (Rule)next ; 
            ActiveRule arule = new ActiveRule(rule,this.tree) ;  
            arule.index = indx ;
            indx++ ;       
            if (rule.existential>0) {   // SHOULD BE!
               if (existential==null) existential = new Vector<ActiveInference>() ;
               existential.add(arule) ;
            }
            else { // arule is definite rule
               if (definite==null) definite = new Vector<ActiveInference>() ;
               definite.add(arule) ; 
            }
         }
         else { // next is a static program rewrite 
            Rewrite rw = (Rewrite)next ; 
            ActiveRewrite arw = new ActiveRewrite(rw,this.tree) ; 
            // ActiveRewrite's index is rw's input location in input theory
            arw.index = indx ; // arw's index is rw's input location in theory
            indx++ ; 
            if (definite==null) definite = new Vector<ActiveInference>() ;       
            definite.add(arw) ; 
            // check for grounded rewrite input
            boolean grounded = (rw.U==0 && rw.V==0) ; // no variables
            if (grounded) { // add equality fact to BRANCH 
               // generate corresponding fact and assert to tree.BRANCH
               Term[] args = {rw.Lterm,rw.Rterm} ;  
               Functor fact = new Functor("=",null,args,(byte)2) ; // Lterm=Rterm 
               fact.index = Theory.EQ_INDEX ; // 3, needed for atpHashCode
               tree.BRANCH[tree.branchNext] = fact ;
               tree.branchNext++ ;
            } 
         }
      }

      // if (definite!=null) definite.trimToSize() ; // reconsider later
      // if (existential!=null) existential.trimToSize() ; // reconsider later
      /* definite is a static component, based on input Theory,  
         but active rule consequents can add new dynamic
         ActiveRewrite(Functor left,Functor right) components to dynamic,  
         to be used in in subsequent inference stages. */
   }

      /**
        * Add a dynamic ActiveRewrite corresponding to tree.BRANCH[indx].
        *    YET TO DO: AVOID REPEATS !!!
         */
      protected void assertDynamicRewrite(int indx, Functor left, Functor right) { 
         ActiveRewrite arw = 
              new ActiveRewrite(indx, left, right, this.tree) ;
         if (this.dynamic == null) // create empty dynamic component list
                this.dynamic = new Vector<ActiveRewrite>() ; 
         this.dynamic.add(arw) ;  // YET TO DO: AVOID REPEATS
      }

   /**
     * Report data for definite and existential ActiveInference(s).
     */   
   public String report() { 
      StringBuffer buff = new StringBuffer() ; 
      if (definite!=null) { // report definite
         buff.append("definite active inferences\n") ; 
         Iterator dit = definite.iterator() ;
         while (dit.hasNext()) {   
            Object next = dit.next() ; 
            if (next instanceof ActiveRule) {
               ActiveRule arule = (ActiveRule)next ;
               buff.append(" "+arule.index +" "+arule.rule.toString()+"\n") ; 
            }
            else {
               ActiveRewrite arw = (ActiveRewrite)next ;
               buff.append(" "+arw.index +" "+arw.rewrite.toString()+"\n") ; 
            }
         } // end while
      } // end definite
      if (existential!=null) { // report existential
         buff.append("existential active rules \n") ; 
         Iterator eit = existential.iterator() ;
         while (eit.hasNext()) {   
            ActiveRule ar = (ActiveRule)(eit.next()) ;
            buff.append(" "+ar.index +" "+ar.rule.toString()+"\n") ;
         } 
      }
      if (dynamic!=null) { // report dynamic
         buff.append("dynamic active rewrites \n") ; 
         Iterator dit = dynamic.iterator() ;
         while (dit.hasNext()) {   
            ActiveRewrite arw = (ActiveRewrite)(dit.next()) ;
            buff.append(" @BRANCH"+ arw.atFact +" "+ arw.Lterm +"="+ arw.Rterm +"\n") ;  
            // ^ tricky to express correctly ...
         } 
      }
      return buff.toString() ;
   } // end report()

   /**
     * preliminary tests ...
      */
   public static void main(String[] args) { 
      // Monday test for indices,  10/23/2023
      // somewhat convoluted references 
      // Provide Theory for inference index inspection ...
      String thry =     // input test
         "true => p(a)(X), p(a)(c).  f(X,Y)=g(Y,X). true=>a∘b=b∘a.  p(a)(X) => goal." ;
      Theory theory = Theory.translate(thry) ; 
      AutoLogAnalyzer.analyze(theory) ;  // theory data computed
      System.out.println("input theory:\n"+theory) ;
      Iterator it = theory.iterator() ; 
      int indx = 1 ;
      while (it.hasNext()) {
         Inference infr = (Inference)(it.next()) ;
         System.out.println(indx +"  "+ infr) ;
         indx++ ; 
      } 
      ActiveTree tree = new ActiveTree(theory,10,5,true) ;
      ActiveTheory activetheory = new ActiveTheory(tree) ;  
      // assert a dynamic active rewrite  re 3d inference of theory
      Rule r3 = (Rule)theory.elementAt(2) ; 
      System.out.println("r3: "+r3) ;
      Functor f = (r3.consequent.elementAt(0).elementAt(0)) ;
      System.out.println("f="+f) ;
      Functor left = (Functor)f.args[0] , right = (Functor)f.args[1] ; 
      System.out.println("left="+left) ;
      System.out.println("right="+right) ;
      // ... continue here
      activetheory.assertDynamicRewrite(22, left, right); // 
      System.out.println("----------") ;
      System.out.println(activetheory.report()) ;
   }

} // end ActiveTheory

/* end notes
   1. How is a dynamic ActiveRewrite added to an ActiveTheory?
      public void assertDynamic(ActiveRewrite) ;
   
   2. Modify ADN/active.DGM re current AT. {maybe the assert method}
   3. A dynamic ActiveRewrite has index i where Activetree branch[i] 
      is the a grounded  equality fact -- index for use in inference reports,
      e.g. proofs.  
   4. Figure out how both definite and dynamic should be concurrently 
      processed.
*/

