
package sam.machine.active ;
import sam.machine.lang.* ;
import java.util.* ;
import java.util.concurrent.* ; 
import sam.gui.search.* ; // AutoLog Viewer
import sam.util.* ;  // Pattern

/**  
  *  ActiveTree -- Skolem Machine active inference management
  *  @version 9/25/2023 
  *  - ActiveTree, ActiveTheory, ActiveRule, ActiveRewrite 
  *    all SHARE underlying instance variable "theory". 
  *  - lshift declared STATIC = 1
  *    atpHashCode(-,-) method declared STATIC
   */
public class ActiveTree implements Runnable {
   public Theory theory ;         // user theory/program
   public boolean predicative ;   // = theory.isPredicative()
   public boolean indexical ;     // = theory.isIndexical && !predicative
   // AutoLog does NOT attempt to process (run) a non-indexical theory.  
   public ActiveTheory activetheory ; // dynamic 
   public Functor[] BRANCH ;
   public int depth ;       // restrict depth of BRANCH
   public int width ;       // restrict width of tree 
   public int branchNext ;  // next open BRANCH index   
   // Depth limit reached when branchNext == BRANCH.length.
   public static int lshift = 1 ;  // STATIC default atpHashCode left shift
   public boolean log ;     // logging flag 
   //////////////// ConcurrentHashMaps //////////////////////////// 
   //  for a predicative theory use
     public ConcurrentHashMap<Integer,Vector<Integer>> FactMap ;  
   // for an indexical theory, use double hashing  
     public ConcurrentHashMap<Integer,Vector<Integer>> ArityMap ; // by arity 
     public ConcurrentHashMap<Integer,Vector<Integer>> PredMap ;  // by predicate
   ////////////////////////////////////////////////////////////////
   // activator processes staged ActiveRule and ActiveRewrite components
   public ExecutorService activator ;  

   /**
     * N.B - theory should be analyzed (e.g., by autolog viewer), 
     * AutoLogAnalyzer.analyze(theory) BEFORE constructing ActiveTree.
     * Construct an ActiveTree ONLY if theory.isIndexical().
     */
   public ActiveTree(Theory theory, 
                     int deep, int wide,
                     boolean logit) { 
      this.theory = theory ; 
      if (this.theory.isIndexical())  { // construction 
         this.predicative = theory.isPredicative() ; // NOTE BENE
         this.indexical = theory.isIndexical() && !this.predicative ; 
         if (theory.size() > deep) this.depth=theory.size() ;
         else this.depth=deep ; // avoid too shallow
         this.width = wide ; 
         this.log = logit ; 
         if (this.predicative)
            this.FactMap = 
              new ConcurrentHashMap<Integer,Vector<Integer>>() ;       
              // methodology ≈ colog14I
         // otherwise ... double hashing re arity and predicate
         if (this.indexical) { 
            this.ArityMap =
              new ConcurrentHashMap<Integer,Vector<Integer>>() ; 
            this.PredMap =
              new ConcurrentHashMap<Integer,Vector<Integer>>() ; 
         }
         this.BRANCH = new Functor[depth+1] ;  
         this.BRANCH[0] = new Functor("true",null,null,(byte)0) ; 
         this.branchNext = 1 ; 
         // can substitute many other Executors 
         this.activator = Executors.newWorkStealingPool() ; 
         this.activetheory = new ActiveTheory(this) ;  
      } // constructed ONLY if theory.isIndexical(), and
        // thus theory is either predicative or indexical.
   }  

   /**
    * Create logging data structures. 
    * AutoLog Viewer always turns on logging.
     */
   public void setLog(boolean logit) { 
       this.log = logit ; 
   }

   /**                   
   *   STATIC atpHashCode calculations for a GROUNDED branch fact
   *   using a particular bound argument key Pattern. 
   *   N.B., args are Functor terms.
   *   Primitive recursive aritmetization for logic terms.
    */
   public static int atpHashCode(Functor f, boolean[] key) {
      int h ; 
      if (f.fnctr != null)   // f parametric
         h = atpHashCode(f.fnctr) ;
      else h = f.index ;     // f named
      for (int i = 0 ; i < f.arity ; i++) // use key pattern
         h = (h << lshift) + (key[i] ? atpHashCode(f.args[i]) : 0) ; 
      return h ;        //   ^ N.B. NOT f.key[i] 
   }
   /** For embedded ground terms  */
   public static int atpHashCode(Term t) { 
      Functor f = (Functor)t ;  
      int hc ; 
      if (f.fnctr != null)   // f parametric
         hc = atpHashCode(f.fnctr) ;
      else hc = f.index ;     // f named
      for (int i = 0 ; i < f.arity ; i++) { // accelerate?
         hc = (hc << lshift) + atpHashCode(f.args[i]) ;
      } 
      return hc ; 
   }

   /*
    * Remember a fact F for predicative theory.
    * Use the key profile Patterns computed
    * by the AutoLogAnalyzer and the BRANCH 
    * location at which the fact is asserted.
    */
   public void rememberP(Functor F, int loc) {
      Vector<Pattern> patterns = 
                    this.theory.keyPatterns[F.index] ;   
      Iterator pattern_iterator = patterns.iterator() ; 
      while(pattern_iterator.hasNext()) { 
         Pattern pat = (Pattern)pattern_iterator.next() ;
         int hc = atpHashCode(F,pat.key) ;
         Vector<Integer> probe = this.FactMap.get(hc) ; 
         if (probe == null) { 
            probe = new Vector<Integer>() ; 
            probe.add(loc) ;
         } 
         else if (!probe.contains(loc)) probe.add(loc) ; 
         FactMap.put(hc, probe) ;
      } // end while
   } // end rememberP - needs tests

   /*
    * Remember a fact F for indexical theory.
    * Use the ArityMap and BRANCH location for F.arity 
    * and use the PredMap and BRANCH location for 
    * the bound predicate (named or F.fnctr) of F.
    */
   public void rememberI(Functor F, int loc) {

   }

   /**
     * Start the prover. E.g., the "run" command in the AutoLog
     * Viewer initiates the following run method in ActiveTree.  
     * Staged inference will use activator.invokeAll(definite) 
     * method on the definite ActiveTheory inference components
     * and then will use inferenceServer.invoke(e) on a chosen
     * existential ActiveTheory inference component e (if any).
      */
   public void run() { 
      if (!theory.isPredicative()) return ; // BAIL
      // UNDER CONSTRUCTION ...
   }

   /*
    * Display atpHashCodes generated for fact F,
    * using theory.keyPatterns[F.index].
     */ 
   public String spyP(Functor F) {
      StringBuffer sbuff = new StringBuffer() ;
      sbuff.append("Report atpHashCode(s) for "+F+":\n") ;
      Vector<Pattern> patterns = 
                    this.theory.keyPatterns[F.index] ;   
      Iterator pattern_iterator = patterns.iterator() ; 
      while(pattern_iterator.hasNext()) { 
         Pattern pat = (Pattern)pattern_iterator.next() ;
         int hc = atpHashCode(F,pat.key) ;
         sbuff.append("   atpHashCode(p(a)(b,c), "+pat+") = " +hc+"\n") ;   
      } // end while
      return sbuff.toString() ;
   } // end spyP

   /**
     * main test stub 3 -- rememberP, predicative
     * more stub test automation  
      */
   public static void main(String[] args) { 
      // Provide Theory for tests ...
      /* copy these lines to sbuff
         true => p(a)(b,c), p(a)(c,d).  // facts [true,true]
         p(a)(X,Y) => goal.             // key [false,false]
         p(a)(b,Y) => false.            // key [true,false]
         p(a)(X,e) => false.            // key [false,true]
      */ 
      StringBuffer sbuff = new StringBuffer() ; 
         sbuff.append(" true => p(a)(b,c), p(a)(c,d). ") ; // 1st line
         sbuff.append(" p(a)(X,Y) => goal. ") ; 
         sbuff.append(" p(a)(b,Y) => false. ") ;
         sbuff.append(" p(a)(X,e) => false. ") ;
      String sthry = sbuff.toString() ; 
      Theory theory = Theory.translate(sthry) ; 
      AutoLogAnalyzer.analyze(theory) ; // theory data computed
      System.out.println("The input theory was analyzed:\n"+theory) ;
      Rule r0 = (Rule)theory.elementAt(0) ; // the consequent facts
      Functor fact1 = ((r0.consequent).elementAt(0)).elementAt(0) ;
      Functor fact2 = ((r0.consequent).elementAt(0)).elementAt(1) ;
      // Report re fact1 and fact2 
      System.out.println("Report for fact1 and fact2:") ;
      System.out.println("fact1= "+fact1) ;
      System.out.println("fact1.index= "+fact1.index) ;
      System.out.println("theory.keyPatterns[fact1.index]="+theory.keyPatterns[fact1.index]) ;
      System.out.println("fact2= "+fact2) ;
      System.out.println("fact2.index= "+fact2.index) ;
      System.out.println("theory.keyPatterns[fact2.index]="+theory.keyPatterns[fact2.index]) ;
      // construct ActiveTree for theory
      System.out.println("✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰") ;
      ActiveTree tree = new ActiveTree(theory,20,1, false) ; // theory was analyzed 
      tree.BRANCH[1] = fact1 ; tree.BRANCH[2] = fact2 ;  // assert first rule consequents
      System.out.println(" ActiveTree (tree) constructed for input theory: \n"+theory) ; 
      System.out.println(" tree.BRANCH.length= "+ tree.BRANCH.length) ;
      System.out.println(" tree.width= "+ tree.width) ;
      System.out.println(" tree.BRANCH[0]= "+ tree.BRANCH[0]) ; // true
      System.out.println(" fact1 = tree.BRANCH[1]= "+ tree.BRANCH[1]) ; // ...
      System.out.println(" fact2 = tree.BRANCH[2]= "+ tree.BRANCH[2]) ; // ...

      // A  -- report FactMap keys corresponding to key patterns
      System.out.println("\nfact1= "+fact1) ;
      System.out.println(tree.spyP(fact1)) ;
      System.out.println("\nfact2= "+fact2) ;
      System.out.println(tree.spyP(fact2)) ;
      System.out.println("\nA> tree.rememberP(tree.BRANCH[1],1)") ;
      tree.rememberP(tree.BRANCH[1],1) ; 
      ConcurrentHashMap.KeySetView<Integer,Vector<Integer>> ksetA = tree.FactMap.keySet() ; // !!!
      System.out.println(" ksetA= "+tree.FactMap.keySet()) ; 
      Iterator Ait =ksetA.iterator() ;
      int Acount = 0 ;
      while (Ait.hasNext()) { 
        int i = (int)Ait.next() ;
        Vector<Integer> locs = tree.FactMap.get(i) ;
        System.out.println("locs"+i+"A= "+locs) ;    //  locsiA= locs   
        Acount++ ;
      } // end while A  // compiles now

      // B  -- report FactMap keys corresponding to key patterns
      System.out.println("\nB> tree.rememberP(tree.BRANCH[2],2)") ;
      tree.rememberP(tree.BRANCH[2],2) ; 
      ConcurrentHashMap.KeySetView<Integer,Vector<Integer>> ksetB = tree.FactMap.keySet() ;  // !!!
      System.out.println(" ksetB= "+tree.FactMap.keySet()) ; // ...
      Iterator Bit =ksetB.iterator() ;
      int Bcount = 0 ;
      while (Bit.hasNext()) { 
        int i = (int)Bit.next() ;
        Vector<Integer> locs = tree.FactMap.get(i) ;
        System.out.println("locs"+i+"B= "+locs) ;    //  locsiB= locs   
        Bcount++ ;
      } // end while B
   } // end main -- seems ok, check details

} // end ActiveTree


/* ✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰
END NOTES 
  -  ActiveTree ≈ colog14I/sam/infer/Tree (analogy)
  -  Public components promote action spying for dev code.
  -  sam.machine.lang.Theory specifies Analyzer data components.
  -  AutoLog only attempts to processes(run) predicative or indexical programs.
  -  Preliminary dev motivation and testing (sam.machine.engine) :
     - sam.machine.engine.ActiveMod.java,       
     - sam.machine.engine.ActiveMod_dev/  {Rewrite Loom Test RLT}
     - sam.machine.engine.TangleTool.java
  -  /active_dev/ActiveTree.java.1  {1st inline tests for main}
  -  /active_dev/ActiveTree.java.2  {2nd inline tests for main}
  -  /active_dev/ActiveTree.java.?
  -  /active_dev/ ... attempt rubric for stub test to load

   ✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰✰
*/
