
package sam.machine.engine ;
 
import sam.machine.lang.* ; 
import java.util.* ;
import java.util.concurrent.* ;
import java.time.Duration;
import java.time.Instant;
      
/**
  *  ActiveMod 
  *     dev tests for ActiveRewrite component
  *     using Java 20 preview:
  *        version 1/1/2022 ...
  *         --release 19/20 --enable-preview < compile switches
  *         --enable-preview              < run switch
  *  Runnable w/internal cache of modulants
  *  @version 10/27/2022 --  Using Java21, no preview switches
  *  Run tests from the JavApp editor, NOT the AUTOLOG app.
   */
public class ActiveMod implements Runnable {
   // Use rewrite to modulate fact
   public Rewrite rewrite ;
   public Functor fact ;
   public Vector<Functor> cache ;
   public Functor left  ; 
   public Term    right ;

   /**
     * Construct an ActiveMod task from string, e.g.:
     * "true => p(f(1,2),f(2,3)). f(X,Y) = X." ; 
     *             ^fact                ^rewrite 
     *                             ^left   ^right 
     * A TangleTool test form.
      */
   public ActiveMod(String modtest) {
      Theory theory = Theory.translate(modtest) ; 
      AutoLogAnalyzer.analyze(theory) ; // indexing
      this.cache = new Vector<Functor>() ;
      this.fact =
         ((((Rule)theory.elementAt(0)).consequent).elementAt(0)).elementAt(0) ;
      this.rewrite = (Rewrite)theory.elementAt(1) ; 
      this.left = (Functor)rewrite.Lterm ; // cast 
      this.right= rewrite.Rterm ; 
   } // end constructor

   /** 
     * Employ TangleTool (✓) static methods  
     * to create modulants.
     */
   public void run() { // perform the modulations
      Vector<Integer> path = new Vector<Integer>() ;
      path.add(1) ;
      Functor[] varbind = new Functor[rewrite.U] ; 
      Vector<Vector<Integer>> locs = new Vector<Vector<Integer>>() ;
      Vector<Functor[]> bindings = new Vector<Functor[]>() ;
      TangleTool.locate(left,rewrite.U,fact,path,locs,bindings) ;  // ✓
      for (int i=0 ; i<locs.size() ; i++) { 
         //Functor lterm = 
         // (Functor)TangleTool.substitute(left,bindings.elementAt(i)) ; // ✓
         Functor rterm = 
            (Functor)TangleTool.substitute(right,bindings.elementAt(i)) ; // ✓ 
         Vector<Integer> rpath = locs.elementAt(i) ; // replacement path
         /////  modulate // ✓
         Functor modulant =(Functor)TangleTool.modulate(fact,0,rpath,rterm) ;
         this.cache.add(modulant) ; 
      }
   } // end run

   ///////////////////////////////////////////////
   /////      Concurrency Executors examples 
   /////    Boilerplate executor code for tests
   ///////////////////////////////////////////////
   /**  -- MAIN --
     * Compare   Executors.newFixedThreadPool(-) 
     *      vs   Executors.newVirtualThreadPerTaskExecutor()
     * running concurrent list of ActiveMOD tasks
      */
   public static void main(String[] args) { 
      ///// create list of independent ActiveMod tasks
      int N = 10_000 ;  // number(edit) of ActiveMod tasks to time
      Vector<ActiveMod> tasks = new Vector<ActiveMod>() ;
      for (int n=0 ; n<N ; n++)
         tasks.add(new ActiveMod("true => p(f(1,2),f(2,3)). f(X,Y) = X."))  ; 
      System.out.println(tasks.size()+" tasks") ;
  
      ///// construct executor, CHOOSE: 
      ExecutorService executor = Executors.newWorkStealingPool() ; 
//      ExecutorService executor = Executors.newFixedThreadPool(10) ; 
//      ExecutorService executor = Executors.newVirtualThreadPerTaskExecutor() ; 
//             -- for release 19/20 (not 21) --enable-preview   ^

      ///// submit all tasks to executor 
      Instant start = Instant.now();
      try{ 
         for(int n=0 ; n<N ; n++) 
            executor.submit((ActiveMod)(tasks.elementAt(n))) ;
      } catch(Exception e) { System.out.println(e) ; }
      
      ///// shutdown and await termination
      executor.shutdown() ;    // needs to finish, so
      try {          // added for virtual thread version
           if (!executor.awaitTermination(60, TimeUnit.SECONDS)) {
               executor.shutdownNow();
          }
      } catch (InterruptedException ex) {
           executor.shutdownNow();
           Thread.currentThread().interrupt();
      }
      
      ///// report 
      Instant finish = Instant.now();
      System.out.println("shutdown " + executor.isShutdown()) ;
      System.out.println("terminated " +executor.isTerminated()) ;
      long timeElapsed = Duration.between(start, finish).toMillis();  //in millis
      System.out.println("Total elapsed time : " + timeElapsed+"ms");
      ///// spy on some task cache
      Vector<Functor> cache = tasks.elementAt(5656).cache ;
      for (int i=0 ; i < cache.size() ; i++) 
         System.out.println("cache 5656@"+i+": "+ cache.elementAt(i)) ;
      int p = Runtime.getRuntime().availableProcessors() ;
      System.out.println(p+" #processors") ;

   } // end main
} // end ActiveMOD


/* end notes
ActiveMod does not locate multiple possible matches on a long branch
stage nor assert new facts -- which is what the 
autolog sam.machine.engine.ActiveRewrite component would also do. 

iMacPro 10_000   36 #processors      (18 core Xeon W)
24ms   Executors.newWorkStealingPool()    WSP
33ms   newFixedThreadPool(10)             FTP(10)
51ms   newVirtualThreadPerTaskExecutor()  VTPT   ~40ms using Java21

For WSP N=1_000_000, Total elapsed time : 767ms  ---  Amazing.
Seems like perhaps WSP takes a few seconds to terminate(?)

MacBookAir 10_000   4 #processors    (2core i5)
51ms  WSP
64ms  FTP(10)
175ms VTPT

MacBookPro 10_000    12 #processors  (6core i7)
19ms  WSP
27ms  FTP(10)
47ms  VTPT


*/
