
import java.util.* ;
import java.util.concurrent.* ;

/**  
  *  InferenceManager   (using CountDownLatch)
  *      9/4/2022 test prototype (started 5/7/2019 at Silent Valley)
  *      - constructs Branch
  *      - constructs ActiveRule vector collection
  *      - constructs ExecutorService rulepool
  *      - invokes rulepool for all active rules
  *      - waits for all ActiveRules to finish (CountDownLatch)
  *      - merges ActiveRule cashes to Branch
  *      - reports stats and branch contents.
   */
public class InferenceManager {   ///// implements Runnable 
   public static void main(String[] args) throws InterruptedException { 
      // construct & run
      int numrules = Integer.parseInt(args[0]) ; // user input
      int stall    = Integer.parseInt(args[1]) ; // user input
      CountDownLatch doneSignal = new CountDownLatch(numrules);
      // set up branch frame
      Branch frame = new Branch() ;  
      // Set up collection of active rules 
      Vector<ActiveRule> activerules = new Vector<ActiveRule>() ;
      //ActiveRule[] activerules = new ActiveRule[numrules] ;
      for (int r=0; r < numrules ; r++) 
         activerules.add(new ActiveRule(r,stall,doneSignal)) ; 
      ExecutorService inferencepool = Executors.newFixedThreadPool(numrules) ;
      // run 
System.out.println("Go!") ;
      // all ActiveRules proceed, mark current time  
      long time = System.currentTimeMillis() ; // begin timing
      List<Future<Boolean>> results = inferencepool.invokeAll(activerules) ;   
      // wait for ActiveRules to finish ...
      doneSignal.await();        // wait for all to finish
System.out.println("\nActiveRules Done!") ;
      //  now merge caches for activerules to branch
      for (int r=0 ; r<numrules ; r++)
         frame.mergeFacts(activerules.elementAt(r).cache) ;
System.out.println("Merge to Branch Done!") ;
      time = System.currentTimeMillis()-time ;
      System.out.print("concurrent: #procs="+Runtime.getRuntime().availableProcessors()) ;
      System.out.print(" #rules="+numrules) ;
      System.out.print(" delay="+stall) ;
      System.out.print(" tcc="+time+"ms ") ;
      if (args.length > 2) {   // use time. from SerialInference
         int tser = Integer.parseInt(args[2]) ; 
         System.out.print("tser="+tser) ;
         System.out.println(" ∆L="+-delta(tser,time)+"%") ;    
      } else System.out.println("\n") ;

System.out.println(frame) ;  // show the branch
      // show the Inference Future(s)
      Iterator it = results.iterator() ;
      while (it.hasNext()) { 
         try {
            System.out.println( ((Future)it.next()).get() );
         }
         catch(ExecutionException ee) {
            System.out.println(ee) ;
         } 
      }
      inferencepool.shutdown() ; 
      // Without shutdown() runtime process still shows w/ "ps -U username"
      // InferenceManager needs to know when to stop and shutdown.
System.out.println("inferencepool.isShutdown()="+inferencepool.isShutdown()+"\n") ; 
   } //end main -- seems to work 
  

   // ∆Latency = (tser-tcc)/tser 
   private static float delta(long tser, long tcc) {
      float I = ((float)tser - (float)tcc)/(float)tser ;
      float N = (float)Runtime.getRuntime().availableProcessors() ; 
      float r = I/N ;
      return Math.round(I*100.0f) ;
   }

}

//-------------------------------------------------------------

class Branch {  // data structure
   protected Vector<Fact> frame ;

   protected Branch() {
      this.frame = new Vector<Fact>() ;
   }

   protected void mergeFacts(Vector<Fact> cache) {
      Iterator it = cache.iterator() ; 
      while (it.hasNext()) { 
         Fact fact = (Fact)it.next() ;
         if (!occurs(fact)) frame.add(fact) ;
           }
        }

   private boolean occurs(Fact f) {
      Iterator it = frame.iterator() ;
      while (it.hasNext()) {
         if (f.data==((Fact)it.next()).data) return true; 
      }
      return false ;
   }

   public String toString() { 
      StringBuffer sb = new StringBuffer() ;
      int w = 10 ; // per line
      Iterator it = this.frame.iterator() ;
      while (it.hasNext()) {       
         if (w==0) { sb.append("\n") ; w=10 ; } // new line  
         sb.append(it.next().toString()) ; w-- ; 
      }
      return sb.toString() ;
   }
} // end Branch

//-------------------------------------------------------------

class ActiveRule implements Callable<Boolean> { 
   //private final CountDownLatch startSignal;
   private final CountDownLatch doneSignal;
   protected int id ;              // this Rule id
   protected Vector<Fact> cache ;  // temporary cache of facts 
   protected int stall ;           // take some time

   protected ActiveRule(int id,int stall, CountDownLatch doneSignal) {
      this.id = id ;
      this.stall = stall ;
      this.cache = new Vector<Fact>() ;  
      //this.startSignal = startSignal;
      this.doneSignal = doneSignal;
   }

   public Boolean call() {   
      try { 
System.out.print(this.id+ ", ") ;
         // add new facts
         for (int k=5*this.id ; k<=5*this.id+9; k++) 
            this.cache.add(new Fact(this.id,k)) ; 
         // stall some 
         Thread.sleep(stall) ; 
         doneSignal.countDown() ; // this ActiveRule is done
      } catch (InterruptedException e) {
           System.out.println(this.id + " interrupted:" + e.toString()) ;
        } 
      return true ;   // return true if cache contains 
                      // 'goal' or' false' terminal fact, else return false
   }
/////////////////////////////////////////////////
//  This version ActiveRule does not consult 
//  branch frame to avoid caching tentative fact. 
//  AutoLog would check for such repeats
//  from previous stages: a cache would not add a
//  Fact already occurring on the branch.
//  But frame.merge(-) does NOT repeat fact.data.
//  This test only prototypes fact production from 1 stage.
/////////////////////////////////////////////////
} // end ActiveRule

//-------------------------------------------------------------

class Fact { 
   protected int id ;     // producing rule id
   protected int data ;   // inferred data

   protected Fact(int id,int data) {
      this.id = id ;
      this.data = data;
   }
   
   public String toString() { 
      return this.id+":"+data+", " ;  // "id:data, "
   }
} // end Fact

/*
The iMP is not any faster than the MBP ... 
Probably using fixed thread pools. 
(Maybe use larger thread pool w/executor 
and retest! ...)
*/

