
package sam.machine.engine ;
import sam.machine.lang.* ; 

/** 
  *  Design and test match(w/bindings) algorithm, 
  *  and substitute method (w/bindings) for parametric Terms.
  *  Started 2/2/2020    (Lake Skinner)
  *  @version 4/22/2020  (Tucalota Springs) 
  * 
   */
public class MatchTest { 
   /**
      * IMPORTANT NOTE RE VARIABLE INDICES for match_table
      *   All variables in a rule are indexed using String name
      *   by AutoLogAnalyzer, including Variable Functors.
      *    Thus, in a rule like 
      *         t(P), P, P(a),  P(a,X) => s(P).   // U=2
      *    all "P"s have the same Variable index value:
      *    referenced via match_table[0]. BUT, they would NOT
      *    have the same Functor index value.
      *    P can thus refer to a term name, a proposition, 
      *    a predicate of arity 1, and a predicate of arity 2,
      *    etc., all (binding) indexical via the first literal   
      *    t(P).
       */
   public Functor[] bindings ;  // store matched variables ;
   public String varindex ;  // name:index of match Variables
                                   
   /**
     *  Construct a test for matching Term(s) fact and 
     *  literal input a "fake" rule from the AutoLog editor:
     *     match, fact, literal => result.  
     *  Match fact and literal using table and then
     *  substitute values of variables into result.    
     *  See line#789  for method match_test() in Viewer sam/guiAutoLog.java
     *  'match', fact, literal => result.
      */
   public MatchTest(Rule r) { 
      this.bindings = new Functor[r.U] ; // in case result has free vars
      this.varindex = r.varindex ;  
   }
      
   /**
     * Match fixed (bound Functor) with flexed (can have vars) using  
     * match_table, return result bound via the matching.
     *    match, fixed, flexed => result.
      */ 
   public Term match(Term fixed, Term flexed, Term result) {
      if (match(fixed,flexed,bindings))       // note 3 arguments
         return substitute(result,bindings) ; // note 2 arguments
      else return null ;
   }

   /**
     * Match fixed with flexed using match_table. 
     * fixed is a bound functor (branch term).
     * flexed is any Term, which freely allows variables.
      */
   public boolean match(Term fixed, Term flexed,Functor[] bindings) {
      if (flexed instanceof Variable) {  
         // "lookup" the value of flexed
         int i = ((Variable)flexed).index ; // use index of Variable
         if (bindings[i] != null) { 
            Functor f = bindings[i] ; 
            if (match(fixed,f,bindings)) return true ; 
            else return false ;
         }
         else {
            bindings[i] = (Functor)fixed ;   // cast needed?
            return true ;
         }
      }
      else {  
         // fixed & flexed must be Functors
         //  < name  fnctr  args >   arity is #args
         Functor fixt  = (Functor)fixed ;  // convenient cast
         Functor flext = (Functor)flexed ;
         if (flext.fnctr == null) { // flext is NAMED  (~colog14I)
            // and so fixt must also be named 
            if (fixt.fnctr != null) return false ; 
            // ELSE: must have same arity 
            if (flext.arity != fixt.arity) return false ; 
            // ELSE: named,named & same arity
            if (fixt.index != flext.index) return false ;  // N.B same index: name/arity
            // ELSE: names/arity match, so check args ...
            for(int i=0; i < fixt.arity ; i++)   // order for literals? 
               // fail fast ...
               if (!match(fixt.args[i],flext.args[i],bindings)) return false ; 
            // ELSE: all args matched, so 
            return true ; 
         } 
         else {  // flext is PARAMETRIC Functor, flext.fnctr!=null
            if (flext.arity == 0) { 
               //  flext should be Variable Functor -- safety check
               if (!(flext.fnctr instanceof Variable)) return false ; 
               Variable v = (Variable)(flext.fnctr) ; 
               Term t = bindings[v.index] ; // lookup v by index
               if (t == null) { // v NOT matched
                  // update match_table
                  bindings[v.index] = fixt ;  // record v by index
               }
               else { // match t with fixt ?
                  if (!match(fixt,t,bindings)) return false ;
                  else return true ;
               }
            } //ELSE: flext is PARAMETRIC Functor & flext.arity > 0   
            if (fixt.arity != flext.arity) return false ; 
            // ELSE:   fixt & flext have same arity and flext PARAMETRIC
               // e.g.   fixt      flext
               //         n(a)     P(X)
               //         n(a)(b)  P(X)  
               //         n(a)(b)  P(X)(Y)
               //      ... 
            if (fixt.fnctr == null) { // fixt is named
               // flext.fnctr must be a Variable in order to match
               if (!(flext.fnctr instanceof Variable)) return false ; 
               // ELSE: flext.fnctr is Variable which must match (.equals(-)) fixt.name
               // e.g., n(-) P(-)
               Variable v = (Variable)(flext.fnctr) ;
               Functor t = (Functor)(bindings[v.index]) ; // lookup v by index
               if (t != null) { // v bound in match_table 
                  if (!t.name.equals(fixt.name)) return false ; // NOT same names 
//?????           if (t.index != fixt.index) return false ;          
               }
               else {
                  bindings[v.index] =          // record v by index
                     new Functor(fixt.name,null,null,fixt.type) ;    // WHAT ABOUT .index ?
                     // The NAME of a Functor fixt.type, w/o args 
                     // Required modification of Functor toString() if type>0.
                        // Later:
                        // Does this force correct type? 
                        // This NAME forces an operator/type upon substitution.  
               }
               // & now match args ...
               for(int i=0; i < fixt.arity ; i++) 
                   // fail fast ...
                   if (!match(fixt.args[i],flext.args[i],bindings)) return false ; 
               return true ;
            } // ELSE: fixt is also PARAMETRIC bound 
              // recurse on fnctr and args
            if (!match(fixt.fnctr,flext.fnctr,bindings)) return false ;
            for(int i=0; i < fixt.arity ; i++) 
               // fail fast ...
               if (!match(fixt.args[i],flext.args[i],bindings)) return false ; 
            return true ;
         }
      }
   } // END match 

   /**
     * Substitute match_table bindings into target. 
     * Provisos:
     * - Care required to force return to correct forms,
     *   PREFIX, INFIX, POSTFIX, DUOP 
     * - Return any unbound Variable as-is in target.
     * - new Functor(-,-,-,-) : what about .index?
     * - ?
      */
   public Term substitute(Term target,Functor[] bindings) { 
      if (target instanceof Variable) {
         int i = ((Variable)target).index ; 
         if (i >= bindings.length) return target ; // not matched
         // e.g., free var in result
         else return bindings[i] ;
      } 
      else { // target is Functor     
         Functor f = (Functor)(target) ; 
         Term[] fargs = null ; 
         if (f.arity>0) { // cart before horse
            fargs = new Term[f.arity] ; 
            for (int i=0 ; i<f.arity ; i++)
               fargs[i]=substitute(f.args[i],bindings) ; 
         }  // otherwise fargs is null
         if (f.fnctr==null)  // NAMED
            return new Functor(f.name,null,fargs,f.type) ;
         else { // f is PARAMETRIC 
            if (f.fnctr instanceof Variable) { 
               // How is the variable bound?
               Term reflect = substitute((Variable)(f.fnctr),bindings) ;
               if (reflect==f.fnctr)  // var was NOT bound in match_table
                  return  // return constructed reflection
                     new Functor(null,f.fnctr,fargs,f.type) ; 
               else { // var reflect was bound to a Functor
                  Functor ref = (Functor)reflect ;
                  if (ref.fnctr == null) // reflect is NAMED Functor
                     return 
                        new Functor (ref.name,null,fargs,ref.type) ;
                  else  // reflect is PARAMETRIC Functor
                     return  
                        new Functor (null,ref.fnctr,fargs,ref.type) ;
              } 
            }
            else { // f.fnctr is a Functor 
               Functor reflect = (Functor) substitute((Functor)(f.fnctr),bindings) ;
               return 
                  new Functor(null,reflect,fargs,f.type) ; // note forcing
            }
         } 
      }   
   }

   /**
     * Inspect bindings.
     * Call AFTER match: mt.report() 
      */
   public String report() { 
      StringBuffer buffer = new StringBuffer() ; 
      buffer.append("Match table indices: \n   ") ; 
      buffer.append(this.varindex +" \n\n") ; 
      buffer.append("bindings: \n") ;
      for (int i = 0 ; i < bindings.length ; i++) { 
         buffer.append("   ") ;
         buffer.append(i + ":" + bindings[i] + "\n") ;
      }
      return buffer.toString() ;
   }

   /**
     * Inspect structure of the target.
     * Call AFTER match: mt.inspect(-) 
      */
   public String inspect(Functor target) { 
      StringBuffer buffer = new StringBuffer() ; 
      buffer.append("Structure of substituted return: \n") ; 
      buffer.append("< -.name="+target.name+", \n") ; 
      buffer.append("  -.fnctr="+target.fnctr+", \n") ; 
      buffer.append("  -.arity="+target.arity+", \n") ; 
      buffer.append("  -.type="+target.type+"   >\n") ; 
      return buffer.toString() ;
   }

   /**
     *  Examine data for substitution casting ...
      */
   public static void main(String[] s) { 
      Term[] args = new Term[0]; 
      Functor f = new Functor("∂",null,args,(byte)0) ;
      Functor g = new Functor("∂",null,args,(byte)2) ;
      System.out.println(f.name==g.name) ; 
   }

   // DEBUG TOOL: notify where code gives a false return
   private boolean falseAT(int n) { 
      System.out.println("false@"+n) ;
      return false ;
   }

}

/*  ENDNOTES

Location of the match test rig in sam.gui.search.AutoLog.java
   -line 374 JMenuItem("match test") 
   -line 828 else if (command.equals("match_test")) match_test() ;
   -line 1116 protected synchronized void match_test() 

>>> issue 1  .type     See BENCHMARK/mt_compare.txt
   p(a), P(a) => P(b).
   ^ --- ^ ----> ^      the .type for P 
>>> issue 2  .index
     .index
   p(a), P(X) => P(X,b).
   ^ --- ^ ----> ^
   substitute p/1 and 
   becomes p(a,b) with index p/2
   which may be a "new" predicate for tabling.
   ActiveTheory needs to use "dynamic tabling".
   At present the predicative literal indexing is static 
   on compilation of the input theory, and there
   is no dynamic updating for asserted facts.
   This needs to be DEVELOPED further -- difficult!
   "New kinds of data get added to a branch and the active
   components must learn about the new indexing in order to 
   put that indexing into effect for subsequent inference 
   stages.
--- issue 3 
   Consider parse/compile issues for DUOPs.
   Consider issues 1 and 2 for DUOPS.
   
*/

