package sam.machine.engine ; 

import java.util.* ;
import java.io.* ; 
import sam.machine.lang.* ;
import sam.util.* ;
import sam.gui.search.* ;
import java.awt.* ;
import javax.swing.* ;  
import java.util.concurrent.*; 

/**    
  *  TangleTool  -- FactMapTest /  ModTest
  *  Tests designed to study how to retrieve branch terms
  *  for rule satisfaction and equality modulation 
  *  for an autolog Skolem Machine theory/program.   
  *  ATP (algebraic term profile) storage and retrieval design testkit
  *  for rules, modified locatiion/descrimination net for active rewrite
  *  modulators. Design/test active components for sam.machine.engine.
  *  @version 7/30/2022  
  *     >>> new constructor w/SHIFT
  *     Autolog VIEWER now has SHIFTspinner selector for SHIFT value
  *  @author John R. Fisher 
   */ 
public class TangleTool {
   /**  
     * Read/compile a branch and a tangle test query from AutoLogEdit
     *   a theory-branch rule, e.g.
     *      true =>               // branch[0]
     *         p(a,b),            // branch[1]   
     *         p(a,f(a)),         // branch[2]    
     *         a=c.               // branch[3]  
     *   a tangle query, e.g. 
     *    % p(X,b) => fact.      // match literal p(X,b) with  branch facts
     *    % f(a)=b.              // modulate branch facts using rewrite f(a)=b
     *    % f(X)=g(X).           // modulate branch facts using rewrite f(X)=g(X)
      */
   protected Theory theory ;   
   // computed by constructor using maxArity(theory), for ALL rules and rewrites
   protected int maxArity ;     
   protected Functor[] BRANCH ;  // constructed from first inference Rule form
   protected Inference query ;   // the second input form (Rule or Rewrite)
   protected static int SHIFT ;  // set by constructor - colog14I used 5
   // key profile source, MAX of 10 args for TangleTool testing
   protected boolean[][] KEY = new boolean[1024][10] ; 

   /* 
      FactMap: ConcurrentHashmap retrieval via algebraic term profiles for branch facts.
      To be documented in Autolog Design Notebook Chapter 6.
   */
   protected ConcurrentHashMap<Integer,Vector<Integer>> FactMap ;  
   
   // power=2^arity         arity: 0 1 2 3  4  5  6  7   8   9   10
   protected static int[] power = {1,2,4,8,16,32,64,128,256,512,1024} ;  
   // accelerator for pre-computation of KEY, called by TangleTool constructor
   // See END NOTE #2.  TangleTool over-fills FactMap in order to test 
   // for map bucket overlaps

   protected void setKEY() {   
      // row 0 is all false
      for (int row=1 ; row < 1024 ; row++)  {
         // increment the key pattern
         boolean carry = true ;  // increment last row
         for (int col=0; col < 10; col++) {
            if (!KEY[row-1][col]) {  // previous row key
               if (carry) {
                  KEY[row][col]=true ;
                  carry=false ;
               }  
               else KEY[row][col]=false ;
            }
            else { // KEY[row-1][col]==true
               if (carry) {  
                  KEY[row][col]=false ;
                  carry=true ;
               } 
               else KEY[row][col]=true ;
            }
         }
      }
   }
  
   /**
     *  Check to see if a theory generates an acceptable test 
     *  before constructing TangleTool: 
     *     see tangle_tool() method in sam.gui.AutoLog.java
     *  (Rewrite query allowed to be not indexical for testing.(
      */
   public static boolean check(Theory theory) { 
      if (theory.isEmpty() || theory.size() != 2 
                           || !(theory.elementAt(0) instanceof Rule)) 
         return false ;
      // branch rule first
      Rule rule = (Rule)(theory.elementAt(0)) ;
      if (rule.antecedent.size()>1)  return false ; 
      if (rule.consequent.size()>1)  return false ; 
      Term t = rule.antecedent.elementAt(0) ;
      if (!t.name.equals("true") | t.arity!=0 )  return false ;    
      Conjunction c = (Conjunction)(rule.consequent.elementAt(0)) ;   
      if (!c.factual)  return false ;  // via AutoLogAnalyzer for theory  
      // query 
      Inference query = (Inference)(theory.elementAt(1)) ;   
      if (query instanceof Rule) {  // check form of rule
         Rule qrule = (Rule)query ; 
         if (!qrule.isIndexical()) return false ;  // query rule must be indexical
         if (qrule.consequent.size() > 1) return false ;
         Conjunction qc = (Conjunction)(qrule.consequent.elementAt(0)) ; 
         if (qc.size() > 1) return false ;
         Term qt = qc.elementAt(0) ;
         if (!qt.name.equals("fact") | qt.arity!=0) return false ;
         if (qrule.antecedent.size() > 1) return false ;
      } 
      // otherwise check that query is an initial Rewrite 
      else if (query instanceof Rewrite) { 
         Rewrite r = (Rewrite)query ;
         if (r.U < r.V) return false ;  // no free variables in Rterm 
      }
      return true ;
   }

   /** 
     *  Construct tool components assuming ✔ checked, 
     *  TangleTool refers to auxArea of Viewer
      */
   public TangleTool(Theory thy, int shift) { 
      // sam.gui.search.Autolog.java has protected theory
      // which is "shared" with this TangleTool after checked:
      this.theory = thy ;  
      this.SHIFT = shift ; // << shift
      this.maxArity = maxArity(this.theory) ;  
      Rule rule = (Rule)this.theory.elementAt(0) ;
      Conjunction ant = (Conjunction)(rule.antecedent) ;
      Conjunction cons = rule.consequent.elementAt(0) ; 
      this.BRANCH = new Functor[1+cons.size()] ; 
      BRANCH[0] = ant.elementAt(0) ;  // true root
      // rest of branch fact terms ...
      for (int i=0; i<cons.size(); i++) 
         this.BRANCH[i+1] = cons.elementAt(i) ; 
      // extract the query
      this.query = (Inference)this.theory.elementAt(1) ; 
      // accelerate the KEY component
      this.setKEY() ;
      this.FactMap = new ConcurrentHashMap<Integer,Vector<Integer>>() ; 
//      this.ModMap = new ConcurrentHashMap<Integer,Vector<Vector<Integer>>>() ; 
      // The default load factor of a HashMap is 0.75f.
   }

   // largest functor arity in a theory
   // Suggest CANCEL tests if maxArity > 10
   // Production runtime could allow larger arity
   // using various to-be-decided adaptations.
   protected int maxArity(Theory thy) { 
      int n=0 ; 
      for(int j=0 ; j < theory.arity.length ; j++) 
         if (thy.arity[j] > n) n = thy.arity[j] ;
      return n ;
   } 

   /**                   
   *   atpHashCode calculations for a GROUNDED branch fact
   *   using a particular bound key pattern. 
   *   N.B., args terms are Functor terms.
   *   Use KEY list as key accelerator in rememberFact.
    */
   public static int atpHashCode(Functor f, boolean[] key) {
      int h ; 
      if (f.fnctr != null)   // f parametric
         h = atpHashCode(f.fnctr) ;
      else h = f.index ;     // f named
      for (int i = 0 ; i < f.arity ; i++) // use key pattern
         h = (h << SHIFT) + (key[i] ? atpHashCode(f.args[i]) : 0) ; 
      return h ;        //   ^ N.B. NOT f.key[i] 
   }
   /** For embedded ground terms  */
   public static int atpHashCode(Term t) { 
      Functor f = (Functor)t ;  
      int hc ; 
      if (f.fnctr != null)   // f parametric
         hc = atpHashCode(f.fnctr) ;
      else hc = f.index ;     // f named
      for (int i = 0 ; i < f.arity ; i++) { // can accelerate?
         hc = (hc << SHIFT) + atpHashCode(f.args[i]) ;
      } 
      return hc ; 
   }

   /** 
     *  Remember a bound facts's branch location (bloc) 
     *  in FactMap. This (over) remembers ALL key profiles,
     *  and is a test for map overlap. Possibility for overlapss   
     *  is reason for code " if (!locations.contains(bloc)) ",
     *  Production remembering may use analysed key profiles 
     *  where appropriate.
      */
   protected void rememberFact(Functor f, int bloc) { 
      int arity = f.arity ; 
      if (arity == 0) { // a named predicate or constant arg
         Vector<Integer> locations = FactMap.get(f.index) ; 
         if (locations == null) {
            Vector<Integer> vec = new Vector<Integer>() ;
            vec.add(bloc) ;
            FactMap.put(f.index, vec) ; 
         }
         else if (!locations.contains(bloc))
               locations.add(bloc) ; 
         else {} // bloc already recorded
         return ; 
      }
      else { // arity ≥ 1  so assert to FactMap by key 
         for (int row=0 ; row < this.power[arity] ; row++) {    
            boolean[] key = this.KEY[row] ;
            int hcode = atpHashCode(f,key) ; 
            Vector<Integer> locations = FactMap.get(hcode) ;
            if (locations == null) { 
               Vector<Integer> vec = new Vector<Integer>() ;
               vec.add(bloc) ;
               FactMap.put(hcode, vec) ; 
            }
            else if (!locations.contains(bloc)) { 
               locations.add(bloc) ;  
            }
            else {} // bloc already recorded
         }
      }
   }   

   /**
     * Describe internal subterm locations for a Functor f.
     * Discriminate all possible subterms by precise location. 
     * Start with path=[b,...] for BRANCH[b] and initialized empty buffer.
     * Buffer generates the location discrimination report.
      */
   public static void describe(Functor f, Vector<Integer> path, StringBuffer buf) { 
      buf.append (f) ; 
      buf.append("  ") ;
      buf.append("@"+path) ;
      buf.append("\n") ; 
      // describe f.fnctr if not null
      if(f.fnctr != null) { 
         // start a new path
         //Vector<Integer> newpath = Vector<Integer>path.clone() ;
         Vector<Integer> newpath = copy(path) ;  // no warnings?
         newpath.add(-1) ;
         describe((Functor)f.fnctr, newpath, buf) ; 
      }
      // describe all f.args
      for (int i=0 ; i < f.arity ; i++) {
         // start a new path
         //Vector<Integer> newpath = Vector<Integer>path.clone() ;
         Vector<Integer> newpath = copy(path) ;
         newpath.add(i) ; 
         describe((Functor)f.args[i], newpath,buf) ;
      }
   }
   /*
   Method describe is a demonstration experiment to express a
   descrimination network representation for all possible subterms for 
   a branch fact. The query needs to be acceptable modulation test. 
   Use describe to help guide design for locate.
      eg. f=p(a)(f(a),g(a)) @branch fact # b
      desc(f):  
         p(a)(f(a),g(a)) @[b]
         p(a)            @[b,-1]
         a               @[b,-1,0]
         f(a)            @[b,0]
         a               @[b,0,0] 
         g(a)            @[b,1]
         a               @[b,1,0] 
      3/18/2022 finally got this to work as intended.
   */ 

   /**   
    * Quick copy of previous term locator to use 
    * for a next locator. .clone() works but warnings persist.
     */
   public static Vector<Integer> copy(Vector<Integer> in) {
      Vector<Integer> out = new Vector<Integer>(in.size()+1) ;
      Iterator it = in.iterator() ;
      while (it.hasNext()) out.add((Integer)it.next()) ;
      return out ;
   } 

   /**  
     *  Locate (acceptible) lterm matches in a branch fact f
     +  using design approach with match                                             
      */
   public static void locate(Functor lterm,                // lterm= -.
                             int numvars,                  // #vars in lterm
                             Functor f,                    // BRANCH[i] 
                             Vector<Integer> path,         // [i....] 
                             Vector<Vector<Integer>> locs, // [[..], ...]
                             Vector<Functor[]> bindings       // [Functor[], ...]
                                                      ){ 
      // are lterm and f matchable? 
      Functor[] varbind = new Functor[numvars] ;  // for this attempt
      if (match(f,lterm,varbind)) { // f is fixed, lterm is flexed, varbind is a binding
         locs.add(path) ; 
         bindings.add(varbind) ;
      }
      // locate lterm in f.fnctr
      if(f.fnctr != null) { 
         // start a new path
         //Vector<Integer> newpath = Vector<Integer>path.clone() ;
         Vector<Integer> newpath = copy(path) ;  // no warnings?
         newpath.add(-1) ;
         locate(lterm, numvars, (Functor)f.fnctr, newpath, locs, bindings) ; 
      }
      // locate lterm in f.args
      for (int i=0 ; i < f.arity ; i++) {
         // start a new path
         //Vector<Integer> newpath = Vector<Integer>path.clone() ;
         Vector<Integer> newpath = copy(path) ;
         newpath.add(i) ;    
         locate(lterm, numvars, (Functor)f.args[i], newpath,locs, bindings) ;
      }
   }

   /** 
     * @version 4/24/2022 allow flexed to be a general parametric Term
     *- Match fixed with flexed using match_table. 
     *- fixed is a bound functor (branch term).
     *- flexed is any Term, which freely allows variables.
     *- This method was originally specified in MatchTest.
      */
   public static boolean match(Term fixed, Term flexed, Functor[] bindings) {
      if (flexed instanceof Variable) {  
         // "lookup" the value of flexed
         int i = ((Variable)flexed).index ; // use index of Variable
         if (bindings[i] != null) { 
            Functor f = bindings[i] ; 
            if (match(fixed,f,bindings)) return true ; 
            else return false ;
         }
         else {
            bindings[i] = (Functor)fixed ;   // cast needed?
            return true ;
         }
      }
      else {  
         // fixed & flexed must be Functors
         //  < name  fnctr  args >   arity is #args
         Functor fixt  = (Functor)fixed ;  // convenient cast
         Functor flext = (Functor)flexed ;
         if (flext.fnctr == null) { // flext is NAMED  (~colog14I)
            // and so fixt must also be named 
            if (fixt.fnctr != null) return false ; 
            // ELSE: must have same arity 
            if (flext.arity != fixt.arity) return false ; 
            // ELSE: named,named & same arity
            if (fixt.index != flext.index) return false ;  // N.B same index: name/arity
            // ELSE: names/arity match, so check args ...
            for(int i=0; i < fixt.arity ; i++)   // order for literals? 
               // fail fast ...
               if (!match(fixt.args[i],flext.args[i],bindings)) return false ; 
            // ELSE: all args matched, so 
            return true ; 
         } 
         else {  // flext is PARAMETRIC Functor, flext.fnctr!=null
            if (flext.arity == 0) { 
               //  flext should be Variable Functor -- safety check
               if (!(flext.fnctr instanceof Variable)) return false ; 
               Variable v = (Variable)(flext.fnctr) ; 
               Term t = bindings[v.index] ; // lookup v by index
               if (t == null) { // v NOT matched
                  // update match_table
                  bindings[v.index] = fixt ;  // record v by index
               }
               else { // match t with fixt ?
                  if (!match(fixt,t,bindings)) return false ;
                  else return true ;
               }
            } //ELSE: flext is PARAMETRIC Functor & flext.arity > 0   
            if (fixt.arity != flext.arity) return false ; 
            // ELSE:   fixt & flext have same arity and flext PARAMETRIC
               // e.g.   fixt      flext
               //         n(a)     P(X)
               //         n(a)(b)  P(X)  
               //         n(a)(b)  P(X)(Y)
               //      ... 
            if (fixt.fnctr == null) { // fixt is named
               // flext.fnctr must be a Variable in order to match
               if (!(flext.fnctr instanceof Variable)) return false ; 
               // ELSE: flext.fnctr is Variable which must match (.equals(-)) fixt.name
               // e.g., n(-) P(-)
               Variable v = (Variable)(flext.fnctr) ;
               Functor t = (Functor)(bindings[v.index]) ; // lookup v by index
               if (t != null) { // v bound in match_table 
                  if (!t.name.equals(fixt.name)) return false ; // NOT same names 
                  // if (t.index != fixt.index) return false ;  // wrong!   
               }
               else {
                  bindings[v.index] =          // record v by index
                     new Functor(fixt.name,null,null,fixt.type) ;    // WHAT ABOUT .index ?
                     // The NAME of a Functor fixt.type, w/o args 
                     // Required modification of Functor toString() if type>0.
                        // Later:
                        // Does this force correct type? 
                        // This NAME forces an operator/type upon substitution.  
               }
               // & now match args ...
               for(int i=0; i < fixt.arity ; i++) 
                   // fail fast ...
                   if (!match(fixt.args[i],flext.args[i],bindings)) return false ; 
               return true ;
            } // ELSE: fixt is also PARAMETRIC bound 
              // recurse on fnctr and args
            if (!match(fixt.fnctr,flext.fnctr,bindings)) return false ;
            for(int i=0; i < fixt.arity ; i++) 
               // fail fast ...
               if (!match(fixt.args[i],flext.args[i],bindings)) return false ; 
            return true ;
         }
      }
   } // END match 

   /**
     * Substitute match bindings into target. 
     * - This method was originally specified in MatchTest.
     *   Provisos:
     * - Care required to force return to correct forms,
     *   PREFIX, INFIX, POSTFIX, DUOP 
     * - Return any unbound Variable as-is in target.
     * - new Functor(-,-,-,-) : what about .index?
      */
   public static Term substitute(Term target,Functor[] bindings) { 
      if (target instanceof Variable) {
         int i = ((Variable)target).index ; 
         if (i >= bindings.length) return target ; // not matched
         // e.g., free var in result
         else return bindings[i] ;
      } 
      else { // target is Functor     
         Functor f = (Functor)(target) ; 
         Term[] fargs = null ; 
         if (f.arity>0) { // cart before horse
            fargs = new Term[f.arity] ; 
            for (int i=0 ; i<f.arity ; i++)
               fargs[i]=substitute(f.args[i],bindings) ; 
         }  // otherwise fargs is null
         if (f.fnctr==null)  // NAMED
            return new Functor(f.name,null,fargs,f.type) ;
         else { // f is PARAMETRIC 
            if (f.fnctr instanceof Variable) { 
               // How is the variable bound?
               Term reflect = substitute((Variable)(f.fnctr),bindings) ;
               if (reflect==f.fnctr)  // var was NOT bound in match_table
                  return  // return constructed reflection
                     new Functor(null,f.fnctr,fargs,f.type) ; 
               else { // var reflect was bound to a Functor
                  Functor ref = (Functor)reflect ;
                  if (ref.fnctr == null) // reflect is NAMED Functor
                     return 
                        new Functor (ref.name,null,fargs,ref.type) ;
                  else  // reflect is PARAMETRIC Functor
                     return  
                        new Functor (null,ref.fnctr,fargs,ref.type) ;
              } 
            }
            else { // f.fnctr is a Functor 
               Functor reflect = (Functor) substitute((Functor)(f.fnctr),bindings) ;
               return 
                  new Functor(null,reflect,fargs,f.type) ; // note forcing
            }
         } 
      }   
   }

   /**
     *  Modulate a branch fact.  
     *  Replace the target term by rterm at path position p using 
     *  Copy structure only along the actual path to replacement
     *     fact at BRANCH[n] is target of modulation
     *     path   [-,-,-,-,-]   
     *     p       0 1 2 3 4    path pointer
     *     move p right until p==path.size()-1, then replacement
     *     path[p] determines where to replace  P(..X..)  
     *                         either parametric^   ^argument
     *   @version 5/5/2022 revision { target.args.clone() }
     */
   public static Term modulate(Functor target,       // to be modulated
                               int p,                // current position in path
                               Vector<Integer> path, // path to modulant subterm 
                               Functor rterm )   {   // replacement lterm=rterm
      if (p == path.size()-1) return rterm ; 
      // otherwise "clopy" current target 
      Functor f =  // NOTE the clone ---------------------->
         new Functor(target.name, target.fnctr, target.args.clone(), target.type) ;  
      int where = path.elementAt(p+1) ; //  next subterm on path
      if (where >= 0) { // replace an argument at where
         f.args[where] = modulate((Functor)(f.args[where]),p+1,path,rterm) ;
      }
      else { // parametric modulation 
         f.fnctr = modulate((Functor)(f.fnctr),p+1,path,rterm) ; 
      }
      return f ; 
   } // end modulate 

//<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
   /** 
     *  showTest()
     *  branch and query setup  
     *  Working with maxArity, 2/27/2022
      */
   public String showTest() {   // branch and query setup
      StringBuffer sb = new StringBuffer() ; 
      sb.append("TangleTool Test Problem \n") ; 
      sb.append("BRANCH \n") ; 
      for (int i=0 ; i < this.BRANCH.length ; i++) {
         sb.append("   "+ i + " ") ;
         sb.append(this.BRANCH[i] +"\n") ;
      }
      // WARN if maxArity>10 -- should CANCEL test. 
      //    TangleTool restricts maxArity≤10, because all 
      //    possible keys are mapped (to test for overlapss).
      //    Production active runtime components may relax 
      //    this testing restriction.
      sb.append("maxArity="+this.maxArity+ (this.maxArity>10 ? " TOO LARGE(CANCEL) \n"  : "\n")) ;
      // Which test?
      sb.append("TANGLE TEST:  \n") ;  
      if (query instanceof Rule) { 
         Term literal = ((Rule)query).antecedent.elementAt(0) ;
         sb.append("Find matches for literal "+literal+" with branch facts.\n") ;
      } 
      else { // query instance of Rewrite
         Rewrite rewrite = (Rewrite)query ; 
         sb.append("Apply rewrite "+ rewrite.Lterm+"="+rewrite.Rterm +" to BRANCH[1].\n") ;
      } 
      return sb.toString() ; 
   }

   /**
     *  LAUNCH TangleTool QUERY for FactMapTest or ModTest
     *  TangleTool query widget 
     *  Viewer calls construction of TangleTool before widget displayed.
      */
   public void showQueryWidget(AutoLog viewer) { 
      // Set up for JOptionPane.showOptionDialog 
      Object[] options = { "OK", "CANCEL" };
      int result = 
         JOptionPane.showOptionDialog(null, "Click OK to continue", "TANGLE TEST",
             JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
             null, options, options[0]);
      if (result == 0) {  // code tests, see END NOTES re early tests ...  
      if (this.query instanceof Rule) 
          FactMapTest(viewer) ;  
      else ModTest(viewer) ; 
      }
   } 

   ////////////////////
   // FactMapTest  inspect FactMap.toString() after remembering branch facts, 
   //    and locate which branch[i]'s relevant for query 'F=>fact.'
   //    using F.key.
   // 
   ////////////////////
   protected void FactMapTest(AutoLog viewer) {   
      String test_statement = this.showTest() ;  
      for (int i=0 ; i < this.BRANCH.length ; i++) 
         rememberFact((Functor)this.BRANCH[i], i) ;  
      // isolate Functor literal, subject of this test
      Functor f = ((Rule)this.query).antecedent.elementAt(0) ;  
      int hashf = atpHashCode(f, f.key) ; // NOTE f.key
      int checksum = 0 ;  // total number of hashcode patterns 
      for (int i=1 ; i<this.BRANCH.length ; i++)
             checksum+= Math.pow(2,this.BRANCH[i].arity) ; 
      // Check whether all hashcode computations yielded a distinct value
      String check = 
             (FactMap.keySet().size()-1 == checksum ? " ✔ (no overlap)" : " (overlap)") ;
      StringBuffer buf = new StringBuffer() ;  
      buf.append(test_statement+"\n\n") ;
      buf.append("atpHashCode SHIFT (<<) value =" + SHIFT + "\n") ;
      buf.append("FactMap mappings ⇲ \n") ;
      buf.append(FactMap.toString()) ; // ;-) so simple 
      buf.append("\n\nFactMap.keyset().size()="+FactMap.keySet().size() + check ) ;
      buf.append("\nQuery result: \n") ;
      buf.append(f +"↦"+hashf+"="+ FactMap.get(hashf)) ;  
      viewer.auxArea.setText(String.valueOf(buf.toString())) ; 
   }   
   
   ////////////////////
   //   ModTest    Under Construction 
   //  Simulate actions for a single ActiveModulator using
   //  subterm locations and variable bindings.
   //  Note AutoLogAnalyzer version 2/9/2022:
   //  Rule keyPatterns and Lterm static rewriteKeyPatterns separated.
   ////////////////////
   protected void ModTest(AutoLog viewer) {   
      String test_statement = this.showTest() ;  
      Rewrite r = (Rewrite)this.query ;
      Functor left = (Functor)r.Lterm ; // cast
      Term    right = r.Rterm ;
      StringBuffer buf = new StringBuffer() ;  
      buf.append(test_statement+"\n") ;
      boolean[] key = left.key ;  
      // ----- describe -----
      buf.append("Describe subterm locations for [1]"+BRANCH[1]+"\n") ; 
      Functor f = BRANCH[1] ; 
      Vector<Integer> path = new Vector<Integer>() ;
      path.add(1) ;
      StringBuffer bfr = new StringBuffer() ;
      describe(f, path, bfr) ;    
      bfr.append("-----------------\n\n") ;
      buf.append(bfr) ;
      ////////////////////////////////////////////////////////////////
      // ----- Locations, bindings and modulants -----
      Functor[] varbind = new Functor[r.U] ; 
      Vector<Vector<Integer>> locs = new Vector<Vector<Integer>>() ;
      Vector<Functor[]> bindings = new Vector<Functor[]>() ;
      locate(left,r.U,f,path,locs,bindings) ;  // void return
      // reports ...
      buf.append("Lterm key profile="+string(key)+"\n") ;  
      buf.append("Rewrite Variable index: "+r.varindex +" \n") ; 
      buf.append("# match locations="+locs.size()+"\n") ;
      buf.append("Locations, bindings and modulants: \n") ;
      for (int i=0 ; i<locs.size() ; i++) {
         Functor lterm = (Functor)substitute(left,bindings.elementAt(i)) ;
         Functor rterm = (Functor)substitute(right,bindings.elementAt(i)) ; 
         Vector<Integer> rpath = locs.elementAt(i) ; // replacement path
         /////  modulation
         Functor modulant =(Functor)modulate(BRANCH[1],0,rpath,rterm) ;
         /////                                         ^ fixed
         buf.append("@"+rpath+":") ;
         buf.append("\n") ;
         buf.append("  "+lterm+" = "+rterm+"\n") ;
         /////  modulation
         buf.append("modulant= "+modulant+"\n") ;
         buf.append("{ BRANCH[1]= "+BRANCH[1]+" }\n") ;
         /////
         buf.append("----------\n");
      }
      if (locs.isEmpty()) buf.append("NO MATCH\n") ;
      viewer.auxArea.setText(String.valueOf(buf.toString())) ; 
      // FIXED: check e.g., m3.txt or tuff4.txt re issue
      //        of restricted substitutions ...
   } // end ModTest

   // Display a varbind table 
   protected static String string(Functor[] vb) { 
      StringBuffer sb = new StringBuffer() ;
      for (int i=0 ; i<vb.length ; i++) 
         sb.append(i + " ⇾ " + vb[i] + "\n") ;
      return sb.toString() ;
   }

   // Show boolean[] profile keys 
   //    e.g., [false,true] for Lterm h(a,Y)
   protected static String string(boolean[] k) { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append("[") ; 
      for(int i = 0 ; i < k.length ; i++) {
         sb.append(k[i]) ; 
         if (i < k.length -1) sb.append(",") ; 
     }
      sb.append("]") ; 
      return sb.toString() ; 
   }

   /** 
     *  Experiment to reveal how << shift works
     *  Java parses (k<<l + m) as k<<(l+m)  a.k.o right assoc
     *  and k<<(l+m) has more shift spread than (k<<l) + m
      */ 
   public static void main (String[] args) { 
      // true => p(a)(c,d).    in AutoLogEdit
      //         ^4^5 ^6^7 
      System.out.println  ( (((4<<5) + 5)<<5 + 6)<<5 + 7 ) ;    // 1115684864
      System.out.println  ( (((((4<<5) + 5)<<5) + 6)<<5)+ 7 ) ; // 136391
      //    (k<<l)        1 23456    6    5   4    3   2    1
   } 

//<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><> 

} // end TangleTool  


