package sam.machine.lang ; 

import java.util.* ; 

/**
  *  Functor
  *  @version 5/25/2023  
  *  Functor implements all literal and functional terms
  *  for grammar AUTOLOG.g4  
   */
public class Functor extends Term {
   public boolean fact ;    // true if a fact: no variables
   public Term fnctr ;      // non-null if implicit or functorial
   public Term[] args ;     // null if no arguments
   public boolean numeric ;  // set by Reader
   // .name, .index, .arity  inherited from Term
                              
   // how to print as function or operator
   public byte type = 0 ;                  //  default type is FUNCTIONAL
   public static byte FUNCTIONAL = 0 ;     //  F(...)  any arity    
   public static byte PREFIX =     1 ;     //  arity=1
   public static byte INFIX =      2 ;     //  arity=2
   public static byte POSTFIX =    3 ;     //  arity=1  avoid 8/24/2020
   public static byte COLLECTION = 4 ;     //  arity=2  { term | term } 
   public static byte PAIR =       5 ;     //  arity=2  (term , term) 

   // for rule literals only, else null set by Analyzer, see SAM.pdf, section 6
   public boolean[] key ;     // table key pattern (args bound or not)  
   public int[] order ;       // satisfy in order
   // In colog14I see SAM, pg.19 re factual, subfactual.
   /** predicative provided fnctr does NOT contain a variable */
   public boolean predicative ;  // set by AutoLogAnalyzer 
   /** indexical provided fnctr variables are left-to-right already bound */
   public boolean indexical ;    // set by AutoLogAnalyzer 
   /** 
     * factual and subfactual same as colog14I,
     * but pending for autolog as of 8/24 ???, 
     *     grep -r "factual" autolog 
     *     grep -r "subfactual" autolog 
      */
   public boolean factual ;      // set by AutoLogAnalyzer  
   public boolean subfactual ;   // set by AutoLogAnalyzer 

   // isIndexical(), isPredicative(), byName()  inherited from Term

   /**
     * Is this Functor bound?
      */
   public boolean bound() { 
      if (this.predicative && this.factual)  
            return true ;
      else return false ;
   }

   /**
     *  Functor constructor.
     *  Adjusts fnctr parses.
      */
   public Functor(String name, Term fnctr, Term[] args, byte type) { 
      this.fnctr = fnctr ;
      this.arity = (args!=null ? args.length : 0) ; 
      // reveal fnctr via toString()
      if (fnctr != null) {  // give .name a "<.fnctr>" value
         if (fnctr instanceof Variable)  
            this.name = ((Variable)fnctr).name ; 
         else {    // code would parse like this
            this.name =  "‹"+fnctr.toString()+"›" ;  
            this.predicative= false ; // reset by AutoLogAnalyzer
            this.indexical= false ;   // reset by AutoLogAnalyzer 
         }
      }
      else { 
         this.name = name ;
         this.predicative = true ;
         this.indexical = true ;
      }
      this.args = args ;  
      this.type = type ;
   } 

   /**
     *  Generate string representation of this functor. 
     *  Avoids dangling infix expressions,   
     *  'stuff' and "stuff" printed w/quotes
      */
   public String toString() { 
      StringBuffer sb = new StringBuffer() ; 
      if (this.type == Functor.FUNCTIONAL) { 
         if (this.fnctr != null) { 
            if (this.fnctr instanceof Functor &&
                  ((Functor)this.fnctr).type == INFIX)
               sb.append('('+this.fnctr.toString()+')') ;
            else sb.append(this.fnctr.toString())  ;
         }
         else sb.append(this.name) ; 
         if(this.arity > 0) { 
            sb.append("(") ; 
            for(int i = 0 ; i < this.arity ; i++) {
               sb.append(this.args[i].toString()) ;
               if (i < this.arity - 1) sb.append(",") ; 
            }
            sb.append(")") ; 
         }
      }  
      // 8/24 AUTOLOG.g4 parametric functor must be FUNCTIONAL
      else if (this.type == Functor.PREFIX) {  
         sb.append(this.name) ; 
         if (args != null && args.length ==1) 
            sb.append(toString(this.args[0])) ; 
         else // debug signal †
         sb.append("<args error for prefix op>") ; 
      }
      else if (this.type == Functor.POSTFIX) {  
         // 8/24 AUTOLOG.g4 has no POSTFIX
         if (args != null && args.length ==1) { 
            sb.append(toString(this.args[0])) ;
            sb.append(this.name) ;  
         }
         else // debug signal †
         sb.append(this.name+"<args error for postfix op>") ;  
      }
      else if (this.type == Functor.COLLECTION) { // {-|-}
         if (args !=null && args.length ==2) { 
            sb.append("{") ;
            sb.append(toString(this.args[0])) ; 
            sb.append("|") ;              // no spaces
            sb.append(toString(this.args[1])) ; 
            sb.append("}") ;
         } 
         else // debug signal†
            sb.append(this.name+"<args error for collection op>") ; 
      }
      else if (this.type == Functor.PAIR) {   // (-,-)
         if (args !=null && args.length ==2) { 
            sb.append("(") ;
            sb.append(toString(this.args[0])) ; 
            sb.append(",") ;              // no spaces
            sb.append(toString(this.args[1])) ; 
            sb.append(")") ;
         } 
         else // debug signal†
            sb.append(this.name+"<args error for collection op>") ; 
      }
      else { // INFIX
         if (args !=null && args.length ==2) { 
            sb.append(toString(this.args[0])) ; 
            sb.append(this.name) ;              // no spaces
            sb.append(toString(this.args[1])) ; 
         } 
         else // debug signal†
            sb.append(this.name+"<args error for infix op>") ; 
      }
      return sb.toString() ;  
   }

 
   /**
     *  Parenthesize embedded toString infix-infix expressions.
      */
   public static String toString(Term a) { 
      if (a instanceof Functor && ((Functor)a).type==Functor.INFIX) { 
         Functor fa = (Functor)a ;
         StringBuffer sb = new StringBuffer() ;
         sb.append("(") ; 
         sb.append(toString(fa.args[0])) ; 
         sb.append(fa.name) ;     // no spaces
         sb.append(toString(fa.args[1])) ; 
         sb.append(")") ;
         return sb.toString() ; 
      } 
      else if (a instanceof Functor && ((Functor)a).type==Functor.PREFIX) { 
         Functor fa = (Functor)a ;
         StringBuffer sb = new StringBuffer() ;
         sb.append("(") ; 
         sb.append(fa.name) ;     // no spaces
         sb.append(toString(fa.args[0])) ; 
         sb.append(")") ;
         return sb.toString() ; 
      } 
      else if (a instanceof Functor && ((Functor)a).type==Functor.POSTFIX) { 
         Functor fa = (Functor)a ;
         StringBuffer sb = new StringBuffer() ;
         sb.append("(") ; 
         sb.append(toString(fa.args[0])) ; 
         sb.append(fa.name) ;     // no spaces
         sb.append(")") ;
         return sb.toString() ; 
      } 
      else return a.toString() ;
   }
      
   /**
     *  equals if prints same ...
     *  Not quick, but not used for search. 
     *  Note bene:   P(a)  is not equals to Q(a)   
      */
   public boolean equals(Object obj) { 
      if (obj instanceof Functor) 
           return  this.toString().equals(((Functor)obj).toString())   ;
      return false ;
   }

   /**
     *  A
     *  Test stub for toString(),
     *  testing what it looks like when 
     *  Functor.type meets wrong number of Functor.args † 
     *  B 
     *  Test stub for f.bound()
     */
   public static void main(String[] args) {    
     /* A
       //   profile  Functor(String name, Term fnctr, Term[] args, byte type) 
      Term a = new Functor("a",null,null,(byte)0) ;
      Term b = new Functor("b",null,null,(byte)0) ;
      Term[] args1 = new Term[1] ;
         args1[0] = a ;
      Term[] args2 = new Term[2] ;
         args2[0]=a ; args2[1]=b ;
      // PREFIX
      Functor f1a = new Functor("λ",null,args1,(byte)1) ;
      Functor f1op = new Functor("λ",null,null,(byte)1) ;
      System.out.println(f1a +"  "+ f1op +"\n") ;
      // INFIX
      Functor f2a = new Functor("→",null,args2,(byte)2) ;
      Functor f2op = new Functor("→",null,null,(byte)2) ;
      System.out.println(f2a +"  "+ f2op +"\n") ;
      // POSTFIX
      Functor f3a = new Functor("ƒ",null,args1,(byte)3) ;
      Functor f3op = new Functor("ƒ",null,null,(byte)3) ;
      System.out.println(f3a +"  "+ f3op +"\n") ;
     */ // end A
      System.out.println("{ Test A commented out. }\n") ;
      //B  ---------- test .bound() -----------------------
      String thy = " p(b,a)(F(c)(a,b)) => goal. " ; //<<<<<<<<<<<
                //      ^test Functor
      Theory theory = Theory.translate(thy) ; 
      AutoLogAnalyzer.analyze(theory) ;
      System.out.println("  Test B") ; 
      Functor f = 
         (((Rule)(theory.elementAt(0))).antecedent).elementAt(0) ;
      System.out.println(" f = "+f) ; 
      System.out.println(" f.bound() = " + f.bound()) ; 
      // end B

   }

}

/*

*/
